from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.int32 import op_unary_negation_int32
from ...fable_library.list import (cons, is_empty as is_empty_1, tail as tail_1, head, FSharpList, empty as empty_1, of_seq, length as length_1, exists, pick, map_indexed)
from ...fable_library.map_util import (add_to_dict, get_item_from_dict, remove_from_dict, try_get_value)
from ...fable_library.option import (value as value_3, bind)
from ...fable_library.range import range_big_int
from ...fable_library.reflection import (TypeInfo, int32_type, class_type, union_type)
from ...fable_library.seq import (max_by, iterate, length, item, iterate_indexed, max, is_empty, to_list, delay, collect, singleton, empty, map, try_find_index, filter)
from ...fable_library.string_ import (to_fail, printf, to_console)
from ...fable_library.types import (Array, Union, to_string)
from ...fable_library.util import (safe_hash, compare_primitives, get_enumerator, dispose, ignore, IEnumerable_1, IEnumerator, to_iterator, equals, compare, max as max_1)
from ...fable_library.types import FSharpRef
from ..Helper.collections_ import (ResizeArray_init, ResizeArray_create, ResizeArray_map, ResizeArray_iteri, List_tryPickAndRemove)
from .composite_cell import CompositeCell
from .composite_column import CompositeColumn
from .composite_header import CompositeHeader

def get_empty_cell_for_header(header: CompositeHeader, colum_cell_option: CompositeCell | None=None) -> CompositeCell:
    match_value: bool = header.IsTermColumn
    if match_value:
        (pattern_matching_result,) = (None,)
        if colum_cell_option is None:
            pattern_matching_result = 0

        elif colum_cell_option.tag == 0:
            pattern_matching_result = 0

        elif colum_cell_option.tag == 2:
            pattern_matching_result = 1

        else: 
            pattern_matching_result = 2

        if pattern_matching_result == 0:
            return CompositeCell.empty_term()

        elif pattern_matching_result == 1:
            return CompositeCell.empty_unitized()

        elif pattern_matching_result == 2:
            raise Exception("[extendBodyCells] This should never happen, IsTermColumn header must be paired with either term or unitized cell.")


    else: 
        return CompositeCell.empty_free_text()



def ensure_cell_hash_in_value_map(value: CompositeCell, value_map: Any) -> int:
    hash_1: int = safe_hash(value) or 0
    if hash_1 in value_map:
        return hash_1

    else: 
        add_to_dict(value_map, hash_1, value)
        return hash_1



def _expr821() -> TypeInfo:
    return union_type("ARCtrl.ArcTableAux.ColumnValueRefs", [], ColumnValueRefs, lambda: [[("Item", int32_type)], [("Item", class_type("System.Collections.Generic.Dictionary`2", [int32_type, int32_type]))]])


class ColumnValueRefs(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Constant", "Sparse"]


ColumnValueRefs_reflection = _expr821

def ColumnValueRefs__get_RowCount(this: ColumnValueRefs) -> Any | None:
    if this.tag == 1:
        def projection(kv: Any, this: Any=this) -> int:
            return kv[0]

        class ObjectExpr822:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        return max_by(projection, this.fields[0], ObjectExpr822())

    else: 
        return None



def ColumnValueRefs__Copy(this: ColumnValueRefs) -> ColumnValueRefs:
    if this.tag == 1:
        d: Any = dict([])
        def action(v: Any, this: Any=this) -> None:
            add_to_dict(d, v[0], v[1])

        iterate(action, this.fields[0])
        return ColumnValueRefs(1, d)

    else: 
        return ColumnValueRefs(0, this.fields[0])



def ColumnValueRefs__ToSparse_Z524259A4(this: ColumnValueRefs, row_count: int) -> ColumnValueRefs:
    if this.tag == 1:
        return ColumnValueRefs(1, this.fields[0])

    else: 
        d: Any = dict([])
        for i in range(0, (row_count - 1) + 1, 1):
            add_to_dict(d, i, this.fields[0])
        return ColumnValueRefs(1, d)



def ColumnValueRefs__AsSparse(this: ColumnValueRefs) -> Any:
    if this.tag == 1:
        return this.fields[0]

    else: 
        raise Exception("Cannot convert a constant column to sparse. Use ToSparse first.")



def ColumnValueRefs__AsConstant(this: ColumnValueRefs) -> int:
    if this.tag == 1:
        raise Exception("Cannot convert a sparse column to constant. Use ToSparse first.")

    else: 
        return this.fields[0]



def ColumnValueRefs_fromCellColumn_14896151(column: Any, previous_row_count: int, value_map: Any) -> ColumnValueRefs:
    l: int = length(column) or 0
    cells: Any = dict([])
    if l == 0:
        return ColumnValueRefs(1, cells)

    else: 
        current: ColumnValueRefs
        hash_1: int = ensure_cell_hash_in_value_map(item(0, column), value_map) or 0
        if l >= previous_row_count:
            current = ColumnValueRefs(0, hash_1)

        else: 
            add_to_dict(cells, 0, hash_1)
            current = ColumnValueRefs(1, cells)

        def action(i: int, cell: CompositeCell, column: Any=column, previous_row_count: Any=previous_row_count, value_map: Any=value_map) -> None:
            nonlocal current
            hash_2: int = ensure_cell_hash_in_value_map(cell, value_map) or 0
            if current.tag == 1:
                cells_2: Any = current.fields[0]
                cells_2[i] = hash_2

            elif current.fields[0] == hash_2:
                cell_hash_1: int = current.fields[0] or 0

            else: 
                cell_hash_2: int = current.fields[0] or 0
                cells_1: Any = dict([])
                for j in range(0, (i - 1) + 1, 1):
                    cells_1[j] = cell_hash_2
                cells_1[i] = hash_2
                current = ColumnValueRefs(1, cells_1)


        iterate_indexed(action, column)
        return current



def ColumnValueRefs__ToCellColumn_Z2609C9DD(this: ColumnValueRefs, value_map: Any, row_count: int, header: CompositeHeader) -> Array[CompositeCell]:
    if this.tag == 1:
        values: Any = this.fields[0]
        class ObjectExpr823:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        default_cell: CompositeCell = get_empty_cell_for_header(header, None) if (len(values) == 0) else get_empty_cell_for_header(header, get_item_from_dict(value_map, get_item_from_dict(values, max(values.keys(), ObjectExpr823()) - 1)))
        def _arrow824(i_1: int, this: Any=this, value_map: Any=value_map, row_count: Any=row_count, header: Any=header) -> CompositeCell:
            return get_item_from_dict(value_map, get_item_from_dict(values, i_1)) if (i_1 in values) else default_cell

        return ResizeArray_init(row_count, _arrow824)

    else: 
        return ResizeArray_create(row_count, get_item_from_dict(value_map, this.fields[0]))



def _expr838() -> TypeInfo:
    return class_type("ARCtrl.ArcTableAux.ArcTableValues", None, ArcTableValues)


class ArcTableValues:
    def __init__(self, cols: Any, value_map: Any, row_count: int) -> None:
        self._columns: Any = cols
        self._valueMap: Any = value_map
        self._rowCount: int = row_count or 0

    @property
    def Columns(self, __unit: None=None) -> Any:
        this: ArcTableValues = self
        return this._columns

    @Columns.setter
    def Columns(self, columns: Any) -> None:
        this: ArcTableValues = self
        this._columns = columns

    @property
    def ColumnCount(self, __unit: None=None) -> int:
        this: ArcTableValues = self
        class ObjectExpr825:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        return 0 if is_empty(this._columns) else (1 + max(this._columns.keys(), ObjectExpr825()))

    @property
    def ValueMap(self, __unit: None=None) -> Any:
        this: ArcTableValues = self
        return this._valueMap

    @ValueMap.setter
    def ValueMap(self, value_map: Any) -> None:
        this: ArcTableValues = self
        this._valueMap = value_map

    @property
    def RowCount(self, __unit: None=None) -> int:
        this: ArcTableValues = self
        return this._rowCount

    @RowCount.setter
    def RowCount(self, row_count: int) -> None:
        this: ArcTableValues = self
        if (row_count < this._rowCount) if (row_count > 0) else False:
            enumerator: Any = get_enumerator(this._columns)
            try: 
                while enumerator.System_Collections_IEnumerator_MoveNext():
                    match_value: ColumnValueRefs = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()[1]
                    if match_value.tag == 1:
                        values: Any = match_value.fields[0]
                        with get_enumerator(to_list(values.keys())) as enumerator_1:
                            while enumerator_1.System_Collections_IEnumerator_MoveNext():
                                key: int = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current() or 0
                                if key >= row_count:
                                    ignore(remove_from_dict(values, key))



            finally: 
                dispose(enumerator)


        this._rowCount = row_count or 0

    @staticmethod
    def from_cell_columns(columns: Array[Array[CompositeCell]]) -> ArcTableValues:
        class ObjectExpr826:
            @property
            def Compare(self) -> Callable[[int, int], int]:
                return compare_primitives

        row_count: int = (0 if (len(columns) == 0) else max(ResizeArray_map(len, columns), ObjectExpr826())) or 0
        value_map: Any = dict([])
        parsed_columns: Any = dict([])
        def action(i: int, column: Array[CompositeCell]) -> None:
            add_to_dict(parsed_columns, i, ColumnValueRefs_fromCellColumn_14896151(column, row_count, value_map))

        iterate_indexed(action, columns)
        return ArcTableValues(parsed_columns, value_map, row_count)

    @staticmethod
    def from_ref_columns(columns: Any, value_map: Any, row_count: int) -> ArcTableValues:
        parsed_columns: Any = dict([])
        def action(i: int, column: ColumnValueRefs) -> None:
            add_to_dict(parsed_columns, i, column)

        iterate_indexed(action, columns)
        return ArcTableValues(parsed_columns, value_map, row_count)

    @staticmethod
    def init(__unit: None=None) -> ArcTableValues:
        value_map: Any = dict([])
        return ArcTableValues(dict([]), value_map, 0)

    def ToCellColumns(self, headers: Array[CompositeHeader]) -> Array[Array[CompositeCell]]:
        this: ArcTableValues = self
        def _arrow827(i: int) -> Array[CompositeCell]:
            header: CompositeHeader = headers[i]
            return ColumnValueRefs__ToCellColumn_Z2609C9DD(get_item_from_dict(this._columns, i), this._valueMap, this._rowCount, header)

        return ResizeArray_init(len(headers), _arrow827)

    def RescanValueMap(self, __unit: None=None) -> None:
        this: ArcTableValues = self
        new_value_map: Any = dict([])
        enumerator: Any = get_enumerator(this._columns)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                kv: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                match_value: ColumnValueRefs = kv[1]
                if match_value.tag == 1:
                    values: Any = match_value.fields[0]
                    enumerator_1: Any = get_enumerator(values)
                    try: 
                        while enumerator_1.System_Collections_IEnumerator_MoveNext():
                            ckv: Any = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current()
                            hash_2: int = ensure_cell_hash_in_value_map(get_item_from_dict(this._valueMap, ckv[1]), new_value_map) or 0
                            values[ckv[0]] = hash_2

                    finally: 
                        dispose(enumerator_1)


                else: 
                    hash_1: int = ensure_cell_hash_in_value_map(get_item_from_dict(this._valueMap, match_value.fields[0]), new_value_map) or 0
                    this._columns[kv[0]] = ColumnValueRefs(0, hash_1)


        finally: 
            dispose(enumerator)

        this._valueMap = new_value_map

    def Copy(self, __unit: None=None) -> ArcTableValues:
        this: ArcTableValues = self
        next_value_map: Any = dict([])
        def action(kv: Any) -> None:
            add_to_dict(next_value_map, kv[0], kv[1].Copy())

        iterate(action, this._valueMap)
        next_columns: Any = dict([])
        def action_1(kv_1: Any) -> None:
            add_to_dict(next_columns, kv_1[0], ColumnValueRefs__Copy(kv_1[1]))

        iterate(action_1, this._columns)
        return ArcTableValues(next_columns, next_value_map, this._rowCount)

    def __hash__(self, __unit: None=None) -> int:
        this: ArcTableValues = self
        hash_1: int = 0
        for i in range(0, (this.ColumnCount - 1) + 1, 1):
            match_value: ColumnValueRefs | None
            pattern_input: tuple[bool, ColumnValueRefs]
            out_arg: ColumnValueRefs = None
            def _arrow828(__unit: None=None) -> ColumnValueRefs:
                return out_arg

            def _arrow829(v: ColumnValueRefs) -> None:
                nonlocal out_arg
                out_arg = v

            pattern_input = (try_get_value(this._columns, i, FSharpRef(_arrow828, _arrow829)), out_arg)
            match_value = pattern_input[1] if pattern_input[0] else None
            if match_value is not None:
                if match_value.tag == 1:
                    values: Any = match_value.fields[0]
                    enumerator: Any = get_enumerator(values)
                    try: 
                        while enumerator.System_Collections_IEnumerator_MoveNext():
                            kv: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                            hash_1 = (((-1640531527 + kv[0]) + (hash_1 << 6)) + (hash_1 >> 2)) or 0
                            hash_1 = (((-1640531527 + kv[1]) + (hash_1 << 6)) + (hash_1 >> 2)) or 0

                    finally: 
                        dispose(enumerator)


                else: 
                    value_hash: int = match_value.fields[0] or 0
                    for i_1 in range(0, (this._rowCount - 1) + 1, 1):
                        hash_1 = (((-1640531527 + i_1) + (hash_1 << 6)) + (hash_1 >> 2)) or 0
                        hash_1 = (((-1640531527 + value_hash) + (hash_1 << 6)) + (hash_1 >> 2)) or 0


        return hash_1

    def __eq__(self, other: Any=None) -> bool:
        this: ArcTableValues = self
        return (safe_hash(this) == safe_hash(other)) if isinstance(other, ArcTableValues) else False

    def Item(self, cr_indices: tuple[int, int]) -> CompositeCell:
        this: ArcTableValues = self
        row: int = cr_indices[1] or 0
        col: int = cr_indices[0] or 0
        if row > (this.RowCount - 1):
            arg_1: int = this.RowCount or 0
            to_fail(printf("Row index %d is out of bounds for ArcTableValues with row count %d."))(row)(arg_1)

        if col > (this.ColumnCount - 1):
            arg_3: int = this.ColumnCount or 0
            to_fail(printf("Column index %d is out of bounds for ArcTableValues with column count %d."))(col)(arg_3)

        match_value: ColumnValueRefs | None
        pattern_input: tuple[bool, ColumnValueRefs]
        out_arg: ColumnValueRefs = None
        def _arrow830(__unit: None=None) -> ColumnValueRefs:
            return out_arg

        def _arrow831(v: ColumnValueRefs) -> None:
            nonlocal out_arg
            out_arg = v

        pattern_input = (try_get_value(this._columns, col, FSharpRef(_arrow830, _arrow831)), out_arg)
        match_value = pattern_input[1] if pattern_input[0] else None
        if match_value is not None:
            col_value_refs: ColumnValueRefs = match_value
            if col_value_refs.tag == 1:
                values: Any = col_value_refs.fields[0]
                return get_item_from_dict(this._valueMap, get_item_from_dict(values, row)) if (row in values) else to_fail(printf("Row value for index %d does not exist in column %d of ArcTableValues."))(row)(col)

            else: 
                return get_item_from_dict(this._valueMap, col_value_refs.fields[0])


        else: 
            return to_fail(printf("Column %d does not exist in ArcTableValues."))(col)


    def GetEnumerator(self, __unit: None=None) -> IEnumerator[Any]:
        this: ArcTableValues = self
        def _arrow837(__unit: None=None) -> IEnumerable_1[Any]:
            def _arrow836(col_i: int) -> IEnumerable_1[Any]:
                match_value: ColumnValueRefs | None
                pattern_input: tuple[bool, ColumnValueRefs]
                out_arg: ColumnValueRefs = None
                def _arrow832(__unit: None=None) -> ColumnValueRefs:
                    return out_arg

                def _arrow833(v: ColumnValueRefs) -> None:
                    nonlocal out_arg
                    out_arg = v

                pattern_input = (try_get_value(this._columns, col_i, FSharpRef(_arrow832, _arrow833)), out_arg)
                match_value = pattern_input[1] if pattern_input[0] else None
                if match_value is not None:
                    match_value_1: ColumnValueRefs = match_value
                    if match_value_1.tag == 1:
                        values: Any = match_value_1.fields[0]
                        def _arrow834(row_i_1: int) -> IEnumerable_1[Any]:
                            return singleton(((col_i, row_i_1), get_item_from_dict(this._valueMap, get_item_from_dict(values, row_i_1)))) if (row_i_1 in values) else empty()

                        return collect(_arrow834, range_big_int(0, 1, this._rowCount - 1))

                    else: 
                        def _arrow835(row_i: int) -> tuple[tuple[int, int], CompositeCell]:
                            return ((col_i, row_i), get_item_from_dict(this._valueMap, match_value_1.fields[0]))

                        return map(_arrow835, range_big_int(0, 1, this._rowCount - 1))


                else: 
                    return empty()


            return collect(_arrow836, range_big_int(0, 1, this.ColumnCount - 1))

        return get_enumerator(delay(_arrow837))

    def __iter__(self) -> IEnumerator[Any]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None=None) -> IEnumerator[Any]:
        this: ArcTableValues = self
        return get_enumerator(this)


ArcTableValues_reflection = _expr838

def ArcTableValues__ctor_39436F7C(cols: Any, value_map: Any, row_count: int) -> ArcTableValues:
    return ArcTableValues(cols, value_map, row_count)


def _007CIsUniqueExistingHeader_007C__007C(existing_headers: IEnumerable_1[CompositeHeader], input: CompositeHeader) -> int | None:
    if ((((input.tag == 3) or (input.tag == 2)) or (input.tag == 1)) or (input.tag == 0)) or (input.tag == 13):
        return None

    elif input.tag == 12:
        def _arrow839(h: CompositeHeader, existing_headers: Any=existing_headers, input: Any=input) -> bool:
            return True if (h.tag == 12) else False

        return try_find_index(_arrow839, existing_headers)

    elif input.tag == 11:
        def _arrow840(h_1: CompositeHeader, existing_headers: Any=existing_headers, input: Any=input) -> bool:
            return True if (h_1.tag == 11) else False

        return try_find_index(_arrow840, existing_headers)

    else: 
        def _arrow841(h_2: CompositeHeader, existing_headers: Any=existing_headers, input: Any=input) -> bool:
            return equals(h_2, input)

        return try_find_index(_arrow841, existing_headers)



def try_find_duplicate_unique(new_header: CompositeHeader, existing_headers: IEnumerable_1[CompositeHeader]) -> int | None:
    active_pattern_result: int | None = _007CIsUniqueExistingHeader_007C__007C(existing_headers, new_header)
    if active_pattern_result is not None:
        index: int = active_pattern_result or 0
        return index

    else: 
        return None



def try_find_duplicate_unique_in_array(existing_headers: IEnumerable_1[CompositeHeader]) -> FSharpList[dict[str, Any]]:
    def loop(i_mut: int, duplicate_list_mut: FSharpList[dict[str, Any]], header_list_mut: FSharpList[CompositeHeader], existing_headers: Any=existing_headers) -> FSharpList[dict[str, Any]]:
        while True:
            (i, duplicate_list, header_list) = (i_mut, duplicate_list_mut, header_list_mut)
            (pattern_matching_result, header, tail) = (None, None, None)
            if is_empty_1(header_list):
                pattern_matching_result = 0

            elif is_empty_1(tail_1(header_list)):
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1
                header = head(header_list)
                tail = tail_1(header_list)

            if pattern_matching_result == 0:
                return duplicate_list

            elif pattern_matching_result == 1:
                has_duplicate: int | None = try_find_duplicate_unique(header, tail)
                i_mut = i + 1
                duplicate_list_mut = cons({
                    "HeaderType": header,
                    "Index1": i,
                    "Index2": value_3(has_duplicate)
                }, duplicate_list) if (has_duplicate is not None) else duplicate_list
                header_list_mut = tail
                continue

            break

    def predicate(x: CompositeHeader, existing_headers: Any=existing_headers) -> bool:
        return not x.IsTermColumn

    return loop(0, empty_1(), of_seq(filter(predicate, existing_headers)))


def SanityChecks_validateColumnIndex(index: int, column_count: int, allow_append: bool) -> None:
    if index < 0:
        raise Exception("Cannot insert CompositeColumn at index < 0.")

    def _arrow842(__unit: None=None, index: Any=index, column_count: Any=column_count, allow_append: Any=allow_append) -> bool:
        x: int = index or 0
        y: int = column_count or 0
        return (compare(x, y) > 0) if allow_append else (compare(x, y) >= 0)

    if _arrow842():
        raise Exception(("Specified index is out of table range! Table contains only " + str(column_count)) + " columns.")



def SanityChecks_validateRowIndex(index: int, row_count: int, allow_append: bool) -> None:
    if index < 0:
        raise Exception("Cannot insert CompositeColumn at index < 0.")

    def _arrow843(__unit: None=None, index: Any=index, row_count: Any=row_count, allow_append: Any=allow_append) -> bool:
        x: int = index or 0
        y: int = row_count or 0
        return (compare(x, y) > 0) if allow_append else (compare(x, y) >= 0)

    if _arrow843():
        raise Exception(("Specified index is out of table range! Table contains only " + str(row_count)) + " rows.")



def SanityChecks_validateColumn(column: CompositeColumn) -> None:
    ignore(column.Validate(True))


def SanityChecks_validateCellColumns(headers: Array[CompositeHeader], columns: Array[Array[CompositeCell]], raise_exception: bool) -> bool:
    is_valid: bool = True
    en: Any = get_enumerator(headers)
    col_index: int = 0
    if (len(columns) != 0) if (len(headers) != len(columns)) else False:
        def _arrow844(message: str, headers: Any=headers, columns: Any=columns, raise_exception: Any=raise_exception) -> None:
            raise Exception(message)

        def _arrow846(__unit: None=None, headers: Any=headers, columns: Any=columns, raise_exception: Any=raise_exception) -> Callable[[str], None]:
            clo: Callable[[str], None] = to_console(printf("%s"))
            def _arrow845(arg: str) -> None:
                clo(arg)

            return _arrow845

        (_arrow844 if raise_exception else _arrow846())(((("Invalid table. Number of headers (" + str(len(headers))) + ") does not match number of columns (") + str(len(columns))) + ").")
        is_valid = False

    while (len(columns) != 0) if (en.System_Collections_IEnumerator_MoveNext() if is_valid else False) else False:
        header: CompositeHeader = en.System_Collections_Generic_IEnumerator_1_get_Current()
        col_en: Any = get_enumerator(columns[col_index])
        col_index = (col_index + 1) or 0
        while col_en.System_Collections_IEnumerator_MoveNext() if is_valid else False:
            cell: CompositeCell = col_en.System_Collections_Generic_IEnumerator_1_get_Current()
            header_is_data: bool = header.IsDataColumn
            header_is_freetext: bool = (not header.IsDataColumn) if (not header.IsTermColumn) else False
            cell_is_not_freetext: bool = not cell.is_free_text
            if (cell_is_not_freetext if (not cell.is_data) else False) if header_is_data else False:
                def _arrow847(message_1: str, headers: Any=headers, columns: Any=columns, raise_exception: Any=raise_exception) -> None:
                    raise Exception(message_1)

                def _arrow849(__unit: None=None, headers: Any=headers, columns: Any=columns, raise_exception: Any=raise_exception) -> Callable[[str], None]:
                    clo_1: Callable[[str], None] = to_console(printf("%s"))
                    def _arrow848(arg_1: str) -> None:
                        clo_1(arg_1)

                    return _arrow848

                (_arrow847 if raise_exception else _arrow849())(((("Invalid combination of header `" + str(header)) + "` and cell `") + str(cell)) + "`. Data header should contain either Data or Freetext cells.")
                is_valid = False

            if cell_is_not_freetext if header_is_freetext else False:
                def _arrow850(message_2: str, headers: Any=headers, columns: Any=columns, raise_exception: Any=raise_exception) -> None:
                    raise Exception(message_2)

                def _arrow852(__unit: None=None, headers: Any=headers, columns: Any=columns, raise_exception: Any=raise_exception) -> Callable[[str], None]:
                    clo_2: Callable[[str], None] = to_console(printf("%s"))
                    def _arrow851(arg_2: str) -> None:
                        clo_2(arg_2)

                    return _arrow851

                (_arrow850 if raise_exception else _arrow852())(((("Invalid combination of header `" + str(header)) + "` and cell `") + str(cell)) + "`. Freetext header should not contain non-freetext cells.")
                is_valid = False

    return is_valid


def SanityChecks_validateArcTableValues(headers: Array[CompositeHeader], values: ArcTableValues, raise_exception: bool) -> bool:
    is_valid: bool = True
    col_index: int = 0
    if (values.ColumnCount != 0) if (len(headers) != values.ColumnCount) else False:
        def _arrow853(message: str, headers: Any=headers, values: Any=values, raise_exception: Any=raise_exception) -> None:
            raise Exception(message)

        def _arrow855(__unit: None=None, headers: Any=headers, values: Any=values, raise_exception: Any=raise_exception) -> Callable[[str], None]:
            clo: Callable[[str], None] = to_console(printf("%s"))
            def _arrow854(arg: str) -> None:
                clo(arg)

            return _arrow854

        (_arrow853 if raise_exception else _arrow855())(((("Invalid table. Number of headers (" + str(len(headers))) + ") does not match number of columns (") + str(values.ColumnCount)) + ").")
        is_valid = False

    if (values.ColumnCount != 0) if (values.RowCount != 0) else False:
        while (col_index < values.ColumnCount) if is_valid else False:
            header: CompositeHeader = headers[col_index]
            match_value: ColumnValueRefs | None
            key: int = col_index or 0
            pattern_input: tuple[bool, ColumnValueRefs]
            out_arg: ColumnValueRefs = None
            def _arrow856(__unit: None=None, headers: Any=headers, values: Any=values, raise_exception: Any=raise_exception) -> ColumnValueRefs:
                return out_arg

            def _arrow857(v: ColumnValueRefs, headers: Any=headers, values: Any=values, raise_exception: Any=raise_exception) -> None:
                nonlocal out_arg
                out_arg = v

            pattern_input = (try_get_value(values.Columns, key, FSharpRef(_arrow856, _arrow857)), out_arg)
            match_value = pattern_input[1] if pattern_input[0] else None
            if match_value is not None:
                if match_value.tag == 1:
                    column: Any = match_value.fields[0]
                    enumerator: Any = get_enumerator(column)
                    try: 
                        while enumerator.System_Collections_IEnumerator_MoveNext():
                            kv: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                            cell_1: CompositeCell = get_item_from_dict(values.ValueMap, kv[1])
                            if not cell_1.ValidateAgainstHeader(header, raise_exception):
                                is_valid = False


                    finally: 
                        dispose(enumerator)


                else: 
                    hash_1: int = match_value.fields[0] or 0
                    cell: CompositeCell = get_item_from_dict(values.ValueMap, hash_1)
                    if not cell.ValidateAgainstHeader(header, raise_exception):
                        is_valid = False



            col_index = (col_index + 1) or 0

    return is_valid


def Unchecked_tryGetCellAt(column: int, row: int, values: ArcTableValues) -> CompositeCell | None:
    def binder(col: ColumnValueRefs, column: Any=column, row: Any=row, values: Any=values) -> CompositeCell | None:
        if col.tag == 1:
            vals: Any = col.fields[0]
            if row in vals:
                return get_item_from_dict(values.ValueMap, get_item_from_dict(vals, row))

            else: 
                return None


        else: 
            return get_item_from_dict(values.ValueMap, col.fields[0])


    def _arrow860(__unit: None=None, column: Any=column, row: Any=row, values: Any=values) -> ColumnValueRefs | None:
        pattern_input: tuple[bool, ColumnValueRefs]
        out_arg: ColumnValueRefs = None
        def _arrow858(__unit: None=None) -> ColumnValueRefs:
            return out_arg

        def _arrow859(v: ColumnValueRefs) -> None:
            nonlocal out_arg
            out_arg = v

        pattern_input = (try_get_value(values.Columns, column, FSharpRef(_arrow858, _arrow859)), out_arg)
        return pattern_input[1] if pattern_input[0] else None

    return bind(binder, _arrow860())


def Unchecked_getCellWithDefault(column: int, row: int, headers: Array[CompositeHeader], values: ArcTableValues) -> CompositeCell:
    if len(headers) <= column:
        arg_1: int = len(headers) or 0
        to_fail(printf("Column index %d is out of bounds for ArcTableValues with column count %d."))(column)(arg_1)

    match_value: ColumnValueRefs | None
    pattern_input: tuple[bool, ColumnValueRefs]
    out_arg: ColumnValueRefs = None
    def _arrow861(__unit: None=None, column: Any=column, row: Any=row, headers: Any=headers, values: Any=values) -> ColumnValueRefs:
        return out_arg

    def _arrow862(v: ColumnValueRefs, column: Any=column, row: Any=row, headers: Any=headers, values: Any=values) -> None:
        nonlocal out_arg
        out_arg = v

    pattern_input = (try_get_value(values.Columns, column, FSharpRef(_arrow861, _arrow862)), out_arg)
    match_value = pattern_input[1] if pattern_input[0] else None
    if match_value is None:
        if row < values.RowCount:
            return CompositeCell.empty_free_text()

        else: 
            return get_empty_cell_for_header(headers[column], None)


    else: 
        col: ColumnValueRefs = match_value
        if col.tag == 1:
            vals: Any = col.fields[0]
            if row in vals:
                return get_item_from_dict(values.ValueMap, get_item_from_dict(vals, row))

            elif len(vals) == 0:
                return get_empty_cell_for_header(headers[column], None)

            else: 
                header_1: CompositeHeader = headers[column]
                class ObjectExpr863:
                    @property
                    def Compare(self) -> Callable[[int, int], int]:
                        return compare_primitives

                max_i: int = max(vals.keys(), ObjectExpr863()) or 0
                return get_empty_cell_for_header(header_1, get_item_from_dict(values.ValueMap, get_item_from_dict(vals, max_i)))


        else: 
            return get_item_from_dict(values.ValueMap, col.fields[0])




def Unchecked_setCellAt(column_index: int, row_index: int, c: CompositeCell, values: ArcTableValues) -> None:
    if (row_index + 1) > values.RowCount:
        values.RowCount = (row_index + 1) or 0

    match_value: ColumnValueRefs | None
    pattern_input: tuple[bool, ColumnValueRefs]
    out_arg: ColumnValueRefs = None
    def _arrow864(__unit: None=None, column_index: Any=column_index, row_index: Any=row_index, c: Any=c, values: Any=values) -> ColumnValueRefs:
        return out_arg

    def _arrow865(v: ColumnValueRefs, column_index: Any=column_index, row_index: Any=row_index, c: Any=c, values: Any=values) -> None:
        nonlocal out_arg
        out_arg = v

    pattern_input = (try_get_value(values.Columns, column_index, FSharpRef(_arrow864, _arrow865)), out_arg)
    match_value = pattern_input[1] if pattern_input[0] else None
    if match_value is not None:
        col: ColumnValueRefs = match_value
        if col.tag == 1:
            key_1: int = row_index or 0
            value: int = ensure_cell_hash_in_value_map(c, values.ValueMap) or 0
            dict_1_1: Any = col.fields[0]
            if key_1 in dict_1_1:
                dict_1_1[key_1] = value

            else: 
                add_to_dict(dict_1_1, key_1, value)


        else: 
            value_hash: int = col.fields[0] or 0
            hash_3: int = ensure_cell_hash_in_value_map(c, values.ValueMap) or 0
            if hash_3 == value_hash:
                pass

            else: 
                d: Any = dict([])
                for i in range(0, (values.RowCount - 1) + 1, 1):
                    if i == row_index:
                        add_to_dict(d, i, hash_3)

                    else: 
                        add_to_dict(d, i, value_hash)

                values.Columns[column_index] = ColumnValueRefs(1, d)



    elif (row_index == 0) if (values.RowCount <= 1) else False:
        hash_1: int = ensure_cell_hash_in_value_map(c, values.ValueMap) or 0
        add_to_dict(values.Columns, column_index, ColumnValueRefs(0, hash_1))

    else: 
        new_column: Any = dict([])
        add_to_dict(new_column, row_index, ensure_cell_hash_in_value_map(c, values.ValueMap))
        add_to_dict(values.Columns, column_index, ColumnValueRefs(1, new_column))



def Unchecked_removeCellAt(column_index: int, row_index: int, values: ArcTableValues) -> None:
    match_value: ColumnValueRefs | None
    pattern_input: tuple[bool, ColumnValueRefs]
    out_arg: ColumnValueRefs = None
    def _arrow866(__unit: None=None, column_index: Any=column_index, row_index: Any=row_index, values: Any=values) -> ColumnValueRefs:
        return out_arg

    def _arrow867(v: ColumnValueRefs, column_index: Any=column_index, row_index: Any=row_index, values: Any=values) -> None:
        nonlocal out_arg
        out_arg = v

    pattern_input = (try_get_value(values.Columns, column_index, FSharpRef(_arrow866, _arrow867)), out_arg)
    match_value = pattern_input[1] if pattern_input[0] else None
    if match_value is not None:
        col: ColumnValueRefs = match_value
        if col.tag == 1:
            vals: Any = col.fields[0]
            if row_index in vals:
                ignore(remove_from_dict(vals, row_index))
                if len(vals) == 0:
                    ignore(remove_from_dict(values.Columns, column_index))



        elif (values.RowCount == 1) if (row_index == 0) else False:
            ignore(remove_from_dict(values.Columns, column_index))




def Unchecked_moveCellTo(from_col: int, from_row: int, to_col: int, to_row: int, values: ArcTableValues) -> None:
    match_value: CompositeCell | None = Unchecked_tryGetCellAt(from_col, from_row, values)
    if match_value is None:
        pass

    else: 
        cell: CompositeCell = match_value
        Unchecked_removeCellAt(from_col, from_row, values)
        Unchecked_setCellAt(to_col, to_row, cell, values)



def Unchecked_removeHeader(index: int, headers: Array[CompositeHeader]) -> None:
    headers.pop(index)


def Unchecked_removeColumnCells(col_index: int, values: ArcTableValues) -> None:
    ignore(remove_from_dict(values.Columns, col_index))


def Unchecked_removeColumnCells_withIndexChange(col_index: int, column_count: int, values: ArcTableValues) -> None:
    if col_index < (column_count - 1):
        cols: Any = dict([])
        def action(kv: Any, col_index: Any=col_index, column_count: Any=column_count, values: Any=values) -> None:
            c_i: int = kv[0] or 0
            if c_i > col_index:
                add_to_dict(cols, c_i - 1, kv[1])

            elif c_i < col_index:
                add_to_dict(cols, c_i, kv[1])


        iterate(action, values.Columns)
        values.Columns = cols

    else: 
        value: None = Unchecked_removeColumnCells(col_index, values)
        ignore(None)



def Unchecked_removeRowCells(row_index: int, column_count: int, values: ArcTableValues) -> None:
    for c in range(0, (column_count - 1) + 1, 1):
        Unchecked_removeCellAt(c, row_index, values)


def Unchecked_removeRowCells_withIndexChange(row_index: int, values: ArcTableValues) -> None:
    if (values.RowCount == 1) if (row_index == 0) else False:
        values.RowCount = 0
        values.Columns = dict([])

    else: 
        def action_1(kv: Any, row_index: Any=row_index, values: Any=values) -> None:
            match_value: ColumnValueRefs = kv[1]
            if match_value.tag == 1:
                vals: Any = match_value.fields[0]
                if row_index < (values.RowCount - 1):
                    col: Any = dict([])
                    def action(kv_1: Any, kv: Any=kv) -> None:
                        r_i: int = kv_1[0] or 0
                        if r_i > row_index:
                            add_to_dict(col, r_i - 1, kv_1[1])

                        elif r_i < row_index:
                            add_to_dict(col, r_i, kv_1[1])


                    iterate(action, vals)
                    values.Columns[kv[0]] = ColumnValueRefs(1, col)

                else: 
                    ignore(remove_from_dict(vals, row_index))



        iterate(action_1, values.Columns)
        values.RowCount = (values.RowCount - 1) or 0



def Unchecked_removeRowRange_withIndexChange(row_start_index: int, row_end_index: int, values: ArcTableValues) -> None:
    if (values.RowCount == row_end_index) if (row_start_index == 0) else False:
        values.RowCount = 0
        values.Columns = dict([])

    else: 
        def action_1(kv: Any, row_start_index: Any=row_start_index, row_end_index: Any=row_end_index, values: Any=values) -> None:
            match_value: ColumnValueRefs = kv[1]
            if match_value.tag == 1:
                vals: Any = match_value.fields[0]
                range: int = ((row_end_index - row_start_index) + 1) or 0
                if row_end_index < (values.RowCount - 1):
                    col: Any = dict([])
                    def action(kv_1: Any, kv: Any=kv) -> None:
                        r_i: int = kv_1[0] or 0
                        if r_i > row_end_index:
                            add_to_dict(col, r_i - range, kv_1[1])

                        elif r_i < row_start_index:
                            add_to_dict(col, r_i, kv_1[1])


                    iterate(action, vals)
                    values.Columns[kv[0]] = ColumnValueRefs(1, col)

                else: 
                    for row_index in range(row_start_index, row_end_index + 1, 1):
                        ignore(remove_from_dict(vals, row_index))



        iterate(action_1, values.Columns)
        values.RowCount = (values.RowCount - ((row_end_index - row_start_index) + 1)) or 0



def Unchecked_moveColumnCellsTo(from_col: int, to_col: int, values: ArcTableValues) -> None:
    match_value: ColumnValueRefs | None
    pattern_input: tuple[bool, ColumnValueRefs]
    out_arg: ColumnValueRefs = None
    def _arrow868(__unit: None=None, from_col: Any=from_col, to_col: Any=to_col, values: Any=values) -> ColumnValueRefs:
        return out_arg

    def _arrow869(v: ColumnValueRefs, from_col: Any=from_col, to_col: Any=to_col, values: Any=values) -> None:
        nonlocal out_arg
        out_arg = v

    pattern_input = (try_get_value(values.Columns, from_col, FSharpRef(_arrow868, _arrow869)), out_arg)
    match_value = pattern_input[1] if pattern_input[0] else None
    if match_value is not None:
        col: ColumnValueRefs = match_value
        Unchecked_removeColumnCells(from_col, values)
        key_1: int = to_col or 0
        value: ColumnValueRefs = col
        dict_1_1: Any = values.Columns
        if key_1 in dict_1_1:
            dict_1_1[key_1] = value

        else: 
            add_to_dict(dict_1_1, key_1, value)




def Unchecked_addRefColumn(new_header: CompositeHeader, new_col: ColumnValueRefs, row_count: int, index: int, force_replace: bool, only_headers: bool, headers: Array[CompositeHeader], values: ArcTableValues) -> None:
    number_of_new_columns: int = 1
    index_1: int = index or 0
    has_duplicate_unique: int | None = try_find_duplicate_unique(new_header, headers)
    if (has_duplicate_unique is not None) if (not force_replace) else False:
        raise Exception(((("Invalid new column `" + str(new_header)) + "`. Table already contains header of the same type on index `") + str(value_3(has_duplicate_unique))) + "`")

    if row_count > values.RowCount:
        values.RowCount = row_count or 0
        enumerator: Any = get_enumerator(values.Columns)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                kv: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                values.Columns[kv[0]] = ColumnValueRefs__ToSparse_Z524259A4(kv[1], values.RowCount)

        finally: 
            dispose(enumerator)


    if (len(values.Columns) == 1) if (force_replace if ((has_duplicate_unique is not None) if (row_count < values.RowCount) else False) else False) else False:
        values.RowCount = row_count or 0

    if has_duplicate_unique is not None:
        number_of_new_columns = 0
        index_1 = value_3(has_duplicate_unique) or 0

    match_value: int = len(headers) or 0
    match_value_1: int = values.RowCount or 0
    start_col_count: int = match_value or 0
    if has_duplicate_unique is not None:
        Unchecked_removeHeader(index_1, headers)

    headers.insert(index_1, new_header)
    if (has_duplicate_unique is None) if (index_1 < start_col_count) else False:
        def _arrow870(x: int, y: int, new_header: Any=new_header, new_col: Any=new_col, row_count: Any=row_count, index: Any=index, force_replace: Any=force_replace, only_headers: Any=only_headers, headers: Any=headers, values: Any=values) -> int:
            return compare_primitives(x, y)

        last_column_index: int = max_1(_arrow870, start_col_count - 1, 0) or 0
        for column_index in range(last_column_index, index_1 - 1, -1):
            Unchecked_moveColumnCellsTo(column_index, column_index + number_of_new_columns, values)

    if not only_headers:
        key: int = index_1 or 0
        value: ColumnValueRefs = new_col
        dict_1: Any = values.Columns
        if key in dict_1:
            dict_1[key] = value

        else: 
            add_to_dict(dict_1, key, value)




def Unchecked_addColumn(new_header: CompositeHeader, new_cells: Any, index: int, force_replace: bool, only_headers: bool, headers: Array[CompositeHeader], values: ArcTableValues) -> None:
    Unchecked_addRefColumn(new_header, ColumnValueRefs_fromCellColumn_14896151(new_cells, values.RowCount, values.ValueMap), length(new_cells), index, force_replace, only_headers, headers, values)


def Unchecked_addColumnFill(new_header: CompositeHeader, new_cell: CompositeCell, index: int, force_replace: bool, only_headers: bool, headers: Array[CompositeHeader], values: ArcTableValues) -> None:
    Unchecked_addRefColumn(new_header, ColumnValueRefs(0, ensure_cell_hash_in_value_map(new_cell, values.ValueMap)), 1, index, force_replace, only_headers, headers, values)


def Unchecked_fillMissingCells(headers: Array[CompositeHeader], values: ArcTableValues) -> None:
    if values.RowCount == 0:
        pass

    else: 
        for i in range(0, (values.ColumnCount - 1) + 1, 1):
            match_value: ColumnValueRefs | None
            pattern_input: tuple[bool, ColumnValueRefs]
            out_arg: ColumnValueRefs = None
            def _arrow871(__unit: None=None, headers: Any=headers, values: Any=values) -> ColumnValueRefs:
                return out_arg

            def _arrow872(v: ColumnValueRefs, headers: Any=headers, values: Any=values) -> None:
                nonlocal out_arg
                out_arg = v

            pattern_input = (try_get_value(values.Columns, i, FSharpRef(_arrow871, _arrow872)), out_arg)
            match_value = pattern_input[1] if pattern_input[0] else None
            if match_value is not None:
                if match_value.tag == 1:
                    vals: Any = match_value.fields[0]
                    if len(vals) < values.RowCount:
                        header: CompositeHeader = headers[i]
                        def _arrow874(__unit: None=None, headers: Any=headers, values: Any=values) -> CompositeCell:
                            class ObjectExpr873:
                                @property
                                def Compare(self) -> Callable[[int, int], int]:
                                    return compare_primitives

                            i_1: int = max(vals.keys(), ObjectExpr873()) or 0
                            return get_empty_cell_for_header(header, get_item_from_dict(values.ValueMap, get_item_from_dict(vals, i_1)))

                        default_hash: int = ensure_cell_hash_in_value_map(get_empty_cell_for_header(header, None) if (len(vals) == 0) else _arrow874(), values.ValueMap) or 0
                        for j in range(len(vals), (values.RowCount - 1) + 1, 1):
                            if not (j in vals):
                                add_to_dict(vals, j, default_hash)




            else: 
                col: ColumnValueRefs = ColumnValueRefs(0, ensure_cell_hash_in_value_map(get_empty_cell_for_header(headers[i], None), values.ValueMap))
                value: None = add_to_dict(values.Columns, i, col)
                ignore(None)




def Unchecked_addEmptyRow(index: int, values: ArcTableValues) -> None:
    if (values.ColumnCount != 0) if (index < values.RowCount) else False:
        enumerator: Any = get_enumerator(values.Columns)
        try: 
            while enumerator.System_Collections_IEnumerator_MoveNext():
                match_value: ColumnValueRefs = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()[1]
                if match_value.tag == 1:
                    dict_1: Any = match_value.fields[0]
                    class ObjectExpr875:
                        @property
                        def Compare(self) -> Callable[[int, int], int]:
                            return compare_primitives

                    max_index: int = max(dict_1.keys(), ObjectExpr875()) or 0
                    for i in range(max_index, ((index + 1) - 1) - 1, -1):
                        if i in dict_1:
                            v: int = get_item_from_dict(dict_1, i) or 0
                            ignore(remove_from_dict(dict_1, i))
                            add_to_dict(dict_1, i + 1, v)



        finally: 
            dispose(enumerator)


    values.RowCount = (values.RowCount + 1) or 0


def Unchecked_addRow(index: int, new_cells: Array[CompositeCell], headers: Array[CompositeHeader], values: ArcTableValues) -> None:
    row_count: int = values.RowCount or 0
    column_count: int = values.ColumnCount or 0
    increase_row_indices: None
    if index < row_count:
        def _arrow876(x: int, y: int, index: Any=index, new_cells: Any=new_cells, headers: Any=headers, values: Any=values) -> int:
            return compare_primitives(x, y)

        last_row_index: int = max_1(_arrow876, row_count - 1, 0) or 0
        for row_index in range(last_row_index, index - 1, -1):
            for column_index in range(0, (column_count - 1) + 1, 1):
                Unchecked_moveCellTo(column_index, row_index, column_index, row_index + 1, values)

    else: 
        increase_row_indices = None

    def f(column_index_1: int, cell: CompositeCell, index: Any=index, new_cells: Any=new_cells, headers: Any=headers, values: Any=values) -> None:
        Unchecked_setCellAt(column_index_1, index, cell, values)

    set_new_cells: None = ResizeArray_iteri(f, new_cells)


def Unchecked_addRows(index: int, new_rows: Array[Array[CompositeCell]], headers: Array[CompositeHeader], values: ArcTableValues) -> None:
    row_count: int = values.RowCount or 0
    column_count: int = values.ColumnCount or 0
    num_new_rows: int = len(new_rows) or 0
    increase_row_indices: None
    if index < row_count:
        def _arrow877(x: int, y: int, index: Any=index, new_rows: Any=new_rows, headers: Any=headers, values: Any=values) -> int:
            return compare_primitives(x, y)

        last_row_index: int = max_1(_arrow877, row_count - 1, 0) or 0
        for row_index in range(last_row_index, index - 1, -1):
            for column_index in range(0, (column_count - 1) + 1, 1):
                Unchecked_moveCellTo(column_index, row_index, column_index, row_index + num_new_rows, values)

    else: 
        increase_row_indices = None

    current_row_index: int = index or 0
    enumerator: Any = get_enumerator(new_rows)
    try: 
        while enumerator.System_Collections_IEnumerator_MoveNext():
            def f(column_index_1: int, cell: CompositeCell) -> None:
                Unchecked_setCellAt(column_index_1, current_row_index, cell, values)

            set_new_cells: None = ResizeArray_iteri(f, enumerator.System_Collections_Generic_IEnumerator_1_get_Current())
            current_row_index = (current_row_index + 1) or 0

    finally: 
        dispose(enumerator)



def Unchecked_compositeHeaderMainColumnEqual(ch1: CompositeHeader, ch2: CompositeHeader) -> bool:
    return to_string(ch1) == to_string(ch2)


def Unchecked_moveColumnTo(from_col: int, to_col: int, headers: Array[CompositeHeader], values: ArcTableValues) -> None:
    pattern_input: tuple[int, int, int] = ((-1, from_col + 1, to_col)) if (from_col < to_col) else ((1, from_col - 1, to_col))
    shift_start: int = pattern_input[1] or 0
    shift_end: int = pattern_input[2] or 0
    shift: int = pattern_input[0] or 0
    header: CompositeHeader = headers[from_col]
    with get_enumerator(to_list(range_big_int(shift_start, op_unary_negation_int32(shift), shift_end))) as enumerator:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            c: int = enumerator.System_Collections_Generic_IEnumerator_1_get_Current() or 0
            headers[c + shift] = headers[c]
    headers[to_col] = header
    col: ColumnValueRefs = get_item_from_dict(values.Columns, from_col)
    with get_enumerator(to_list(range_big_int(shift_start, op_unary_negation_int32(shift), shift_end))) as enumerator_1:
        while enumerator_1.System_Collections_IEnumerator_MoveNext():
            c_1: int = enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current() or 0
            values.Columns[c_1 + shift] = get_item_from_dict(values.Columns, c_1)
    values.Columns[to_col] = col


def Unchecked_alignByHeaders(keep_order: bool, rows: FSharpList[FSharpList[tuple[CompositeHeader, CompositeCell]]]) -> tuple[Array[CompositeHeader], ArcTableValues]:
    headers: Array[CompositeHeader] = []
    values: ArcTableValues = ArcTableValues.init()
    values.RowCount = length_1(rows) or 0
    def loop(col_i_mut: int, rows_2_mut: FSharpList[FSharpList[tuple[CompositeHeader, CompositeCell]]], keep_order: Any=keep_order, rows: Any=rows) -> tuple[Array[CompositeHeader], ArcTableValues]:
        while True:
            (col_i, rows_2) = (col_i_mut, rows_2_mut)
            def _arrow878(arg: FSharpList[tuple[CompositeHeader, CompositeCell]], col_i: Any=col_i, rows_2: Any=rows_2) -> bool:
                return not is_empty_1(arg)

            if not exists(_arrow878, rows_2):
                return (headers, values)

            else: 
                def _arrow879(l: FSharpList[tuple[CompositeHeader, CompositeCell]], col_i: Any=col_i, rows_2: Any=rows_2) -> tuple[CompositeHeader, CompositeCell] | None:
                    return None if is_empty_1(l) else head(l)

                first_elem: CompositeHeader = pick(_arrow879, rows_2)[0]
                (headers.append(first_elem))
                col_i_mut = col_i + 1
                def mapping(row_i: int, l_1: FSharpList[tuple[CompositeHeader, CompositeCell]], col_i: Any=col_i, rows_2: Any=rows_2) -> FSharpList[tuple[CompositeHeader, CompositeCell]]:
                    if keep_order:
                        if not is_empty_1(l_1):
                            if Unchecked_compositeHeaderMainColumnEqual(head(l_1)[0], first_elem):
                                Unchecked_setCellAt(col_i, row_i, head(l_1)[1], values)
                                return tail_1(l_1)

                            else: 
                                return l_1


                        else: 
                            return empty_1()


                    else: 
                        def f(tupled_arg: tuple[CompositeHeader, CompositeCell], row_i: Any=row_i, l_1: Any=l_1) -> CompositeCell | None:
                            if Unchecked_compositeHeaderMainColumnEqual(tupled_arg[0], first_elem):
                                return tupled_arg[1]

                            else: 
                                return None


                        pattern_input: tuple[CompositeCell | None, FSharpList[tuple[CompositeHeader, CompositeCell]]] = List_tryPickAndRemove(f, l_1)
                        new_l: FSharpList[tuple[CompositeHeader, CompositeCell]] = pattern_input[1]
                        first_match: CompositeCell | None = pattern_input[0]
                        if first_match is None:
                            return new_l

                        else: 
                            Unchecked_setCellAt(col_i, row_i, first_match, values)
                            return new_l



                rows_2_mut = map_indexed(mapping, rows_2)
                continue

            break

    return loop(0, rows)


__all__ = ["get_empty_cell_for_header", "ensure_cell_hash_in_value_map", "ColumnValueRefs_reflection", "ColumnValueRefs__get_RowCount", "ColumnValueRefs__Copy", "ColumnValueRefs__ToSparse_Z524259A4", "ColumnValueRefs__AsSparse", "ColumnValueRefs__AsConstant", "ColumnValueRefs_fromCellColumn_14896151", "ColumnValueRefs__ToCellColumn_Z2609C9DD", "ArcTableValues_reflection", "_007CIsUniqueExistingHeader_007C__007C", "try_find_duplicate_unique", "try_find_duplicate_unique_in_array", "SanityChecks_validateColumnIndex", "SanityChecks_validateRowIndex", "SanityChecks_validateColumn", "SanityChecks_validateCellColumns", "SanityChecks_validateArcTableValues", "Unchecked_tryGetCellAt", "Unchecked_getCellWithDefault", "Unchecked_setCellAt", "Unchecked_removeCellAt", "Unchecked_moveCellTo", "Unchecked_removeHeader", "Unchecked_removeColumnCells", "Unchecked_removeColumnCells_withIndexChange", "Unchecked_removeRowCells", "Unchecked_removeRowCells_withIndexChange", "Unchecked_removeRowRange_withIndexChange", "Unchecked_moveColumnCellsTo", "Unchecked_addRefColumn", "Unchecked_addColumn", "Unchecked_addColumnFill", "Unchecked_fillMissingCells", "Unchecked_addEmptyRow", "Unchecked_addRow", "Unchecked_addRows", "Unchecked_compositeHeaderMainColumnEqual", "Unchecked_moveColumnTo", "Unchecked_alignByHeaders"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (length, min_by, max_by, map, exists)
from ..fable_library.util import (IEnumerable_1, compare_primitives, ignore, get_enumerator, IEnumerator, to_iterator)
from .Cells.fs_cell import FsCell
from .Cells.fs_cells_collection import FsCellsCollection
from .fs_address import FsAddress
from .Ranges.fs_range_address import FsRangeAddress
from .Ranges.fs_range_base import (FsRangeBase, FsRangeBase_reflection)

def _expr225() -> TypeInfo:
    return class_type("FsSpreadsheet.FsRow", None, FsRow, FsRangeBase_reflection())


class FsRow(FsRangeBase):
    def __init__(self, range_address: FsRangeAddress, cells: FsCellsCollection) -> None:
        super().__init__(range_address)
        self.cells_004021: FsCellsCollection = cells

    @staticmethod
    def empty(__unit: None=None) -> FsRow:
        return FsRow(FsRangeAddress(FsAddress(0, 0), FsAddress(0, 0)), FsCellsCollection())

    @staticmethod
    def create_at(index: int, cells: FsCellsCollection) -> FsRow:
        def get_index_by(f: Callable[[Callable[[FsCell], int], IEnumerable_1[FsCell]], FsCell]) -> int:
            if length(cells.GetCellsInRow(index)) == 0:
                return 1

            else: 
                def _arrow218(c: FsCell, f: Any=f) -> int:
                    return c.Address.ColumnNumber

                return f(_arrow218)(cells.GetCellsInRow(index)).Address.ColumnNumber


        def _arrow221(projection: Callable[[FsCell], int]) -> Callable[[IEnumerable_1[FsCell]], FsCell]:
            def _arrow220(source_1: IEnumerable_1[FsCell]) -> FsCell:
                class ObjectExpr219:
                    @property
                    def Compare(self) -> Callable[[int, int], int]:
                        return compare_primitives

                return min_by(projection, source_1, ObjectExpr219())

            return _arrow220

        min_col_index: int = get_index_by(_arrow221) or 0
        def _arrow224(projection_1: Callable[[FsCell], int]) -> Callable[[IEnumerable_1[FsCell]], FsCell]:
            def _arrow223(source_2: IEnumerable_1[FsCell]) -> FsCell:
                class ObjectExpr222:
                    @property
                    def Compare(self) -> Callable[[int, int], int]:
                        return compare_primitives

                return max_by(projection_1, source_2, ObjectExpr222())

            return _arrow223

        max_col_index: int = get_index_by(_arrow224) or 0
        return FsRow(FsRangeAddress(FsAddress(index, min_col_index), FsAddress(index, max_col_index)), cells)

    @property
    def Cells(self, __unit: None=None) -> IEnumerable_1[FsCell]:
        this: FsRow = self
        return super().Cells(this.cells_004021)

    @property
    def Index(self, __unit: None=None) -> int:
        this: FsRow = self
        return this.RangeAddress.FirstAddress.RowNumber

    @Index.setter
    def Index(self, i: int) -> None:
        this: FsRow = self
        this.RangeAddress.FirstAddress.RowNumber = i or 0
        this.RangeAddress.LastAddress.RowNumber = i or 0

    @property
    def MinColIndex(self, __unit: None=None) -> int:
        this: FsRow = self
        return this.RangeAddress.FirstAddress.ColumnNumber

    @property
    def MaxColIndex(self, __unit: None=None) -> int:
        this: FsRow = self
        return this.RangeAddress.LastAddress.ColumnNumber

    def Copy(self, __unit: None=None) -> FsRow:
        this: FsRow = self
        ra: FsRangeAddress = this.RangeAddress.Copy()
        def mapping(c: FsCell) -> FsCell:
            return c.Copy()

        cells: IEnumerable_1[FsCell] = map(mapping, this.Cells)
        fcc: FsCellsCollection = FsCellsCollection()
        fcc.AddMany(cells)
        return FsRow(ra, fcc)

    @staticmethod
    def copy(row: FsRow) -> FsRow:
        return row.Copy()

    @staticmethod
    def get_index(row: FsRow) -> int:
        return row.Index

    def HasCellAt(self, col_index: int) -> bool:
        this: FsRow = self
        def predicate(c: FsCell) -> bool:
            return c.ColumnNumber == col_index

        return exists(predicate, this.Cells)

    @staticmethod
    def has_cell_at(col_index: int, row: FsRow) -> bool:
        return row.HasCellAt(col_index)

    def Item(self, column_index: int) -> FsCell:
        this: FsRow = self
        return super().Cell(FsAddress(1, column_index), this.cells_004021)

    @staticmethod
    def item(col_index: int, row: FsRow) -> FsCell:
        return row.Item(col_index)

    def TryItem(self, col_index: int) -> FsCell | None:
        this: FsRow = self
        return this.Item(col_index) if this.HasCellAt(col_index) else None

    @staticmethod
    def try_item(col_index: int, row: FsRow) -> FsCell | None:
        return row.TryItem(col_index)

    def InsertValueAt(self, col_index: int, value: Any) -> None:
        this: FsRow = self
        cell: FsCell = FsCell(value)
        this.cells_004021.Add(cell, this.Index, col_index)

    @staticmethod
    def insert_value_at(col_index: int, value: Any, row: FsRow) -> FsRow:
        value_1: None = row.InsertValueAt(col_index, value)
        ignore(None)
        return row

    def ToDenseRow(self, __unit: None=None) -> None:
        this: FsRow = self
        for i in range(this.MinColIndex, this.MaxColIndex + 1, 1):
            ignore(this.Item(i))

    @staticmethod
    def to_dense_row(row: FsRow) -> FsRow:
        row.ToDenseRow()
        return row

    @staticmethod
    def create_dense_row_of(row: FsRow) -> FsRow:
        new_row: FsRow = row.Copy()
        new_row.ToDenseRow()
        return new_row

    def GetEnumerator(self, __unit: None=None) -> IEnumerator[FsCell]:
        this: FsRow = self
        return get_enumerator(this.Cells)

    def __iter__(self) -> IEnumerator[FsCell]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None=None) -> IEnumerator[Any]:
        this: FsRow = self
        return get_enumerator(this)


FsRow_reflection = _expr225

def FsRow__ctor_7678C70A(range_address: FsRangeAddress, cells: FsCellsCollection) -> FsRow:
    return FsRow(range_address, cells)


__all__ = ["FsRow_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..fable_library.list import (singleton, FSharpList)
from ..fable_library.map_util import add_to_dict
from ..fable_library.seq import map
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

_VALUE = TypeVar("_VALUE")

def encode_id(id: str) -> IEncodable:
    class ObjectExpr1947(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], id: Any=id) -> Any:
            return helpers.encode_string(id)

    values: FSharpList[tuple[str, IEncodable]] = singleton(("@id", ObjectExpr1947()))
    class ObjectExpr1948(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], id: Any=id) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr1948()


def encode(gen_id: Callable[[_VALUE], str], encoder: Callable[[_VALUE], IEncodable], value: Any, table: Any) -> IEncodable:
    id: str = gen_id(value)
    if id in table:
        return encode_id(id)

    else: 
        v: IEncodable = encoder(value)
        add_to_dict(table, gen_id(value), v)
        return v



__all__ = ["encode_id", "encode"]




from __future__ import annotations
from typing import Any
from ..fable_library.reflection import (TypeInfo, union_type)
from ..fable_library.types import (Array, Union)

def _expr702() -> TypeInfo:
    return union_type("ARCtrl.DataFile", [], DataFile, lambda: [[], [], []])


class DataFile(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["RawDataFile", "DerivedDataFile", "ImageFile"]


DataFile_reflection = _expr702

def DataFile_get_RawDataFileJson(__unit: None=None) -> str:
    return "Raw Data File"


def DataFile_get_DerivedDataFileJson(__unit: None=None) -> str:
    return "Derived Data File"


def DataFile_get_ImageFileJson(__unit: None=None) -> str:
    return "Image File"


def DataFile__get_AsString(this: DataFile) -> str:
    if this.tag == 1:
        return "Derived Data File"

    elif this.tag == 2:
        return "Image File"

    else: 
        return "Raw Data File"



def DataFile_fromString_Z721C83C5(dt: str) -> DataFile:
    (pattern_matching_result,) = (None,)
    if dt == "RawDataFileJson":
        pattern_matching_result = 0

    elif dt == "Raw Data File":
        pattern_matching_result = 0

    elif dt == "DerivedDataFileJson":
        pattern_matching_result = 1

    elif dt == "Derived Data File":
        pattern_matching_result = 1

    elif dt == "ImageFileJson":
        pattern_matching_result = 2

    elif dt == "Image File":
        pattern_matching_result = 2

    else: 
        pattern_matching_result = 3

    if pattern_matching_result == 0:
        return DataFile(0)

    elif pattern_matching_result == 1:
        return DataFile(1)

    elif pattern_matching_result == 2:
        return DataFile(2)

    elif pattern_matching_result == 3:
        raise Exception(("Invalid DataFile type: " + dt) + "")



def DataFile__get_IsDerivedData(this: DataFile) -> bool:
    if this.tag == 1:
        return True

    else: 
        return False



def DataFile__get_IsRawData(this: DataFile) -> bool:
    if this.tag == 0:
        return True

    else: 
        return False



def DataFile__get_IsImage(this: DataFile) -> bool:
    if this.tag == 2:
        return True

    else: 
        return False



__all__ = ["DataFile_reflection", "DataFile_get_RawDataFileJson", "DataFile_get_DerivedDataFileJson", "DataFile_get_ImageFileJson", "DataFile__get_AsString", "DataFile_fromString_Z721C83C5", "DataFile__get_IsDerivedData", "DataFile__get_IsRawData", "DataFile__get_IsImage"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.comment import Comment
from ...arctrl_core.Process.process import Process
from ...arctrl_core.Process.process_input import ProcessInput
from ...arctrl_core.Process.process_output import ProcessOutput
from ...arctrl_core.Process.process_parameter_value import ProcessParameterValue
from ...arctrl_core.Process.protocol import Protocol
from ...arctrl_core.uri import URIModule_toString
from ...fable_library.list import (choose, singleton, of_array, FSharpList)
from ...fable_library.option import map
from ...fable_library.seq import map as map_1
from ...fable_library.string_ import replace
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IOptionalGetter, string, list_1 as list_1_2, IGetters)
from ...thoth_json_core.encode import list_1 as list_1_1
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..comment import (ROCrate_encoder as ROCrate_encoder_5, ROCrate_decoder as ROCrate_decoder_5, ISAJson_encoder as ISAJson_encoder_5, ISAJson_decoder as ISAJson_decoder_5)
from ..context.rocrate.isa_process_context import context_jsonvalue
from ..decode import Decode_uri
from ..encode import (try_include, try_include_list_opt)
from ..idtable import encode
from ..person import (ROCrate_encodeAuthorListString, ROCrate_decodeAuthorListString)
from .process_input import (ROCrate_encoder as ROCrate_encoder_3, ROCrate_decoder as ROCrate_decoder_3, ISAJson_encoder as ISAJson_encoder_3, ISAJson_decoder as ISAJson_decoder_3)
from .process_output import (ROCrate_encoder as ROCrate_encoder_4, ROCrate_decoder as ROCrate_decoder_4, ISAJson_encoder as ISAJson_encoder_4, ISAJson_decoder as ISAJson_decoder_4)
from .process_parameter_value import (ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_2)
from .protocol import (ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_1)

__A_ = TypeVar("__A_")

def ROCrate_genID(p: Process) -> str:
    match_value: str | None = p.ID
    if match_value is None:
        match_value_1: str | None = p.Name
        if match_value_1 is None:
            return "#EmptyProcess"

        else: 
            return "#Process_" + replace(match_value_1, " ", "_")


    else: 
        return URIModule_toString(match_value)



def ROCrate_encoder(study_name: str | None, assay_name: str | None, oa: Process) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2683(__unit: None=None, study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> IEncodable:
        value: str = ROCrate_genID(oa)
        class ObjectExpr2682(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2682()

    class ObjectExpr2684(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> Any:
            return helpers_1.encode_string("Process")

    def _arrow2686(value_2: str, study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> IEncodable:
        class ObjectExpr2685(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr2685()

    def _arrow2687(oa_1: Protocol, study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_1(study_name, assay_name, oa.Name, oa_1)

    def _arrow2688(author_list: str, study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> IEncodable:
        return ROCrate_encodeAuthorListString(author_list)

    def _arrow2690(value_4: str, study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> IEncodable:
        class ObjectExpr2689(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_4)

        return ObjectExpr2689()

    def _arrow2691(value_6: ProcessInput, study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_3(value_6)

    def _arrow2692(value_7: ProcessOutput, study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_4(value_7)

    def _arrow2693(comment: Comment, study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_5(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2683()), ("@type", list_1_1(singleton(ObjectExpr2684()))), try_include("name", _arrow2686, oa.Name), try_include("executesProtocol", _arrow2687, oa.ExecutesProtocol), try_include_list_opt("parameterValues", ROCrate_encoder_2, oa.ParameterValues), try_include("performer", _arrow2688, oa.Performer), try_include("date", _arrow2690, oa.Date), try_include_list_opt("inputs", _arrow2691, oa.Inputs), try_include_list_opt("outputs", _arrow2692, oa.Outputs), try_include_list_opt("comments", _arrow2693, oa.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr2694(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], study_name: Any=study_name, assay_name: Any=assay_name, oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2694()


def _arrow2704(get: IGetters) -> Process:
    def _arrow2695(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("@id", Decode_uri)

    def _arrow2696(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("name", string)

    def _arrow2697(__unit: None=None) -> Protocol | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("executesProtocol", ROCrate_decoder_1)

    def _arrow2698(__unit: None=None) -> FSharpList[ProcessParameterValue] | None:
        arg_7: Decoder_1[FSharpList[ProcessParameterValue]] = list_1_2(ROCrate_decoder_2)
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("parameterValues", arg_7)

    def _arrow2699(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("performer", ROCrate_decodeAuthorListString)

    def _arrow2700(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("date", string)

    def _arrow2701(__unit: None=None) -> FSharpList[ProcessInput] | None:
        arg_13: Decoder_1[FSharpList[ProcessInput]] = list_1_2(ROCrate_decoder_3)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("inputs", arg_13)

    def _arrow2702(__unit: None=None) -> FSharpList[ProcessOutput] | None:
        arg_15: Decoder_1[FSharpList[ProcessOutput]] = list_1_2(ROCrate_decoder_4)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("outputs", arg_15)

    def _arrow2703(__unit: None=None) -> FSharpList[Comment] | None:
        arg_17: Decoder_1[FSharpList[Comment]] = list_1_2(ROCrate_decoder_5)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("comments", arg_17)

    return Process(_arrow2695(), _arrow2696(), _arrow2697(), _arrow2698(), _arrow2699(), _arrow2700(), None, None, _arrow2701(), _arrow2702(), _arrow2703())


ROCrate_decoder: Decoder_1[Process] = object(_arrow2704)

def ISAJson_encoder(study_name: str | None, assay_name: str | None, id_map: Any | None, oa: Process) -> IEncodable:
    def f(oa_1: Process, study_name: Any=study_name, assay_name: Any=assay_name, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], oa_1: Any=oa_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow2708(value: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2707(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2707()

        def _arrow2710(value_2: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2709(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_2)

            return ObjectExpr2709()

        def _arrow2711(oa_2: Protocol, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_1(study_name, assay_name, oa_1.Name, id_map, oa_2)

        def _arrow2712(oa_3: ProcessParameterValue, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_2(id_map, oa_3)

        def _arrow2714(value_4: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2713(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_2.encode_string(value_4)

            return ObjectExpr2713()

        def _arrow2716(value_6: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2715(IEncodable):
                def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_3.encode_string(value_6)

            return ObjectExpr2715()

        def _arrow2717(oa_4: Process, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder(study_name, assay_name, id_map, oa_4)

        def _arrow2718(oa_5: Process, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder(study_name, assay_name, id_map, oa_5)

        def _arrow2719(value_8: ProcessInput, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_3(id_map, value_8)

        def _arrow2720(value_9: ProcessOutput, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_4(id_map, value_9)

        def _arrow2721(comment: Comment, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_5(id_map, comment)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2708, ROCrate_genID(oa_1)), try_include("name", _arrow2710, oa_1.Name), try_include("executesProtocol", _arrow2711, oa_1.ExecutesProtocol), try_include_list_opt("parameterValues", _arrow2712, oa_1.ParameterValues), try_include("performer", _arrow2714, oa_1.Performer), try_include("date", _arrow2716, oa_1.Date), try_include("previousProcess", _arrow2717, oa_1.PreviousProcess), try_include("nextProcess", _arrow2718, oa_1.NextProcess), try_include_list_opt("inputs", _arrow2719, oa_1.Inputs), try_include_list_opt("outputs", _arrow2720, oa_1.Outputs), try_include_list_opt("comments", _arrow2721, oa_1.Comments)]))
        class ObjectExpr2722(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any], oa_1: Any=oa_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_4.encode_object(arg)

        return ObjectExpr2722()

    if id_map is not None:
        def _arrow2723(p: Process, study_name: Any=study_name, assay_name: Any=assay_name, id_map: Any=id_map, oa: Any=oa) -> str:
            return ROCrate_genID(p)

        return encode(_arrow2723, f, oa, id_map)

    else: 
        return f(oa)



def _arrow2736(__unit: None=None) -> Decoder_1[Process]:
    def decode(__unit: None=None) -> Decoder_1[Process]:
        def _arrow2735(get: IGetters) -> Process:
            def _arrow2724(__unit: None=None) -> str | None:
                object_arg: IOptionalGetter = get.Optional
                return object_arg.Field("@id", Decode_uri)

            def _arrow2725(__unit: None=None) -> str | None:
                object_arg_1: IOptionalGetter = get.Optional
                return object_arg_1.Field("name", string)

            def _arrow2726(__unit: None=None) -> Protocol | None:
                object_arg_2: IOptionalGetter = get.Optional
                return object_arg_2.Field("executesProtocol", ISAJson_decoder_1)

            def _arrow2727(__unit: None=None) -> FSharpList[ProcessParameterValue] | None:
                arg_7: Decoder_1[FSharpList[ProcessParameterValue]] = list_1_2(ISAJson_decoder_2)
                object_arg_3: IOptionalGetter = get.Optional
                return object_arg_3.Field("parameterValues", arg_7)

            def _arrow2728(__unit: None=None) -> str | None:
                object_arg_4: IOptionalGetter = get.Optional
                return object_arg_4.Field("performer", string)

            def _arrow2729(__unit: None=None) -> str | None:
                object_arg_5: IOptionalGetter = get.Optional
                return object_arg_5.Field("date", string)

            def _arrow2730(__unit: None=None) -> Process | None:
                arg_13: Decoder_1[Process] = decode(None)
                object_arg_6: IOptionalGetter = get.Optional
                return object_arg_6.Field("previousProcess", arg_13)

            def _arrow2731(__unit: None=None) -> Process | None:
                arg_15: Decoder_1[Process] = decode(None)
                object_arg_7: IOptionalGetter = get.Optional
                return object_arg_7.Field("nextProcess", arg_15)

            def _arrow2732(__unit: None=None) -> FSharpList[ProcessInput] | None:
                arg_17: Decoder_1[FSharpList[ProcessInput]] = list_1_2(ISAJson_decoder_3)
                object_arg_8: IOptionalGetter = get.Optional
                return object_arg_8.Field("inputs", arg_17)

            def _arrow2733(__unit: None=None) -> FSharpList[ProcessOutput] | None:
                arg_19: Decoder_1[FSharpList[ProcessOutput]] = list_1_2(ISAJson_decoder_4)
                object_arg_9: IOptionalGetter = get.Optional
                return object_arg_9.Field("outputs", arg_19)

            def _arrow2734(__unit: None=None) -> FSharpList[Comment] | None:
                arg_21: Decoder_1[FSharpList[Comment]] = list_1_2(ISAJson_decoder_5)
                object_arg_10: IOptionalGetter = get.Optional
                return object_arg_10.Field("comments", arg_21)

            return Process(_arrow2724(), _arrow2725(), _arrow2726(), _arrow2727(), _arrow2728(), _arrow2729(), _arrow2730(), _arrow2731(), _arrow2732(), _arrow2733(), _arrow2734())

        return object(_arrow2735)

    return decode(None)


ISAJson_decoder: Decoder_1[Process] = _arrow2736()

__all__ = ["ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.list import (empty, FSharpList, map, append)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.string_ import (to_text, printf)
from .types import (SheetEntity_1_some_2B595, SheetEntity_1, WorkbookElement, Message__MapText_11D407F6, Message, SheetEntity_1__get_Messages)

def _expr320() -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.WorkbookBuilder", None, WorkbookBuilder)


class WorkbookBuilder:
    def __init__(self, __unit: None=None) -> None:
        pass


WorkbookBuilder_reflection = _expr320

def WorkbookBuilder__ctor(__unit: None=None) -> WorkbookBuilder:
    return WorkbookBuilder(__unit)


def WorkbookBuilder_get_Empty(__unit: None=None) -> SheetEntity_1[FSharpList[WorkbookElement]]:
    return SheetEntity_1_some_2B595(empty())


def WorkbookBuilder__SignMessages_3F89FFFC(this: WorkbookBuilder, messages: FSharpList[Message]) -> FSharpList[Message]:
    def mapping(m: Message, this: Any=this, messages: Any=messages) -> Message:
        def _arrow321(__unit: None=None, m: Any=m) -> Callable[[str], str]:
            clo: Callable[[str], str] = to_text(printf("In Workbook: %s"))
            return clo

        return Message__MapText_11D407F6(m, _arrow321())

    return map(mapping, messages)


def WorkbookBuilder__Combine_49CE2EC0(this: WorkbookBuilder, wx1: SheetEntity_1[FSharpList[WorkbookElement]], wx2: SheetEntity_1[FSharpList[WorkbookElement]]) -> SheetEntity_1[FSharpList[WorkbookElement]]:
    (pattern_matching_result, l1, l2, messages1, messages2, messages2_1, messages1_1, f1, messages1_2, messages2_2, f2, messages1_3, messages2_3, messages1_4, messages2_4) = (None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)
    if wx1.tag == 2:
        if wx2.tag == 2:
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]

        else: 
            pattern_matching_result = 2
            messages1_1 = wx1.fields[0]


    elif wx1.tag == 1:
        if wx2.tag == 0:
            pattern_matching_result = 4
            f2 = wx2.fields[0]
            messages1_3 = wx1.fields[0]
            messages2_3 = wx2.fields[1]

        elif wx2.tag == 1:
            pattern_matching_result = 5
            messages1_4 = wx1.fields[0]
            messages2_4 = wx2.fields[0]

        else: 
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]


    elif wx2.tag == 2:
        pattern_matching_result = 1
        messages2_1 = wx2.fields[0]

    elif wx2.tag == 1:
        pattern_matching_result = 3
        f1 = wx1.fields[0]
        messages1_2 = wx1.fields[1]
        messages2_2 = wx2.fields[0]

    else: 
        pattern_matching_result = 0
        l1 = wx1.fields[0]
        l2 = wx2.fields[0]
        messages1 = wx1.fields[1]
        messages2 = wx2.fields[1]

    if pattern_matching_result == 0:
        return SheetEntity_1(0, append(l1, l2), append(messages1, messages2))

    elif pattern_matching_result == 1:
        return SheetEntity_1(2, append(SheetEntity_1__get_Messages(wx1), messages2_1))

    elif pattern_matching_result == 2:
        return SheetEntity_1(2, append(messages1_1, SheetEntity_1__get_Messages(wx2)))

    elif pattern_matching_result == 3:
        return SheetEntity_1(0, f1, append(messages1_2, messages2_2))

    elif pattern_matching_result == 4:
        return SheetEntity_1(0, f2, append(messages1_3, messages2_3))

    elif pattern_matching_result == 5:
        return SheetEntity_1(1, append(messages1_4, messages2_4))



__all__ = ["WorkbookBuilder_reflection", "WorkbookBuilder_get_Empty", "WorkbookBuilder__SignMessages_3F89FFFC", "WorkbookBuilder__Combine_49CE2EC0"]




from __future__ import annotations

from enum import IntEnum
from urllib.parse import ParseResult, unquote, urljoin, urlparse

from .types import FSharpRef


class UriKind(IntEnum):
    RelativeOrAbsolute = 0
    Absolute = 1
    Relative = 2


class Uri:
    __slots__ = "res", "kind", "orginal"

    def __init__(
        self,
        base_uri: Uri | str,
        kind_or_uri: int | (str | (Uri | None)) = UriKind.Absolute,
    ) -> None:
        self.res: ParseResult

        kind = kind_or_uri if isinstance(kind_or_uri, int) else UriKind.Absolute
        uri = urlparse(base_uri) if isinstance(base_uri, str) else base_uri.res
        relative_uri = (
            urlparse(kind_or_uri)
            if isinstance(kind_or_uri, str)
            else kind_or_uri.res
            if isinstance(kind_or_uri, Uri)
            else None
        )

        # Absolute base URI must be absolute
        if kind == UriKind.Absolute and not uri.netloc:
            raise ValueError("Invalid URI: The format of the URI could not be determined.")

        # Relative base URI must be relative
        if kind == UriKind.Relative and uri.netloc:
            raise ValueError("Invalid URI: The format of the URI could not be determined.")

        if relative_uri:
            self.res = urlparse(urljoin(uri.geturl(), relative_uri.geturl()))
        else:
            self.res = uri
        self.kind = kind
        self.orginal = base_uri

    @property
    def is_absolute_uri(self) -> bool:
        return True if self.res.netloc else False

    @property
    def scheme(self) -> str:
        return self.res.scheme

    @property
    def host(self) -> str:
        return self.res.hostname or ""

    @property
    def absolute_uri(self) -> str:
        if self.kind == UriKind.Relative:
            raise ValueError("This operation is not supported for a relative URI.")

        return self.res.geturl()

    @property
    def absolute_path(self) -> str:
        return self.res.path

    @property
    def query(self) -> str:
        return f"?{self.res.query}"

    @property
    def path_and_query(self) -> str:
        return f"{self.res.path}?{self.res.query}"

    @property
    def fragment(self) -> str:
        return f"#{self.res.fragment}"

    @property
    def original_string(self) -> str:
        return str(self.orginal)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Uri):
            return False

        return self.res == other.res

    def __str__(self) -> str:
        res = self.res._replace(netloc=self.res.netloc.replace(":80", ""))
        return unquote(res.geturl())

    @staticmethod
    def create(uri: str | Uri, kind_or_uri: UriKind | str | Uri = UriKind.Absolute) -> Uri:
        return Uri(uri, kind_or_uri)

    @staticmethod
    def try_create(uri: str | Uri, kind_or_uri: UriKind | str | Uri, out: FSharpRef[Uri]) -> bool:
        try:
            out.contents = Uri.create(uri, kind_or_uri)
            return True
        except Exception:
            return False



from __future__ import annotations
from typing import Any
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .comment import LDComment
from .defined_term import LDDefinedTerm
from .person import LDPerson
from .property_value import LDPropertyValue

def _expr1507() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDScholarlyArticle", None, LDScholarlyArticle)


class LDScholarlyArticle:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/ScholarlyArticle"

    @staticmethod
    def headline() -> str:
        return "http://schema.org/headline"

    @staticmethod
    def identifier() -> str:
        return "http://schema.org/identifier"

    @staticmethod
    def author() -> str:
        return "http://schema.org/author"

    @staticmethod
    def url() -> str:
        return "http://schema.org/url"

    @staticmethod
    def same_as() -> str:
        return "http://schema.org/sameAs"

    @staticmethod
    def creative_work_status() -> str:
        return "http://schema.org/creativeWorkStatus"

    @staticmethod
    def comment() -> str:
        return "http://schema.org/comment"

    @staticmethod
    def try_get_headline_as_string(s: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDScholarlyArticle.headline(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def get_headline_as_string(s: LDNode, context: LDContext | None=None) -> str:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDScholarlyArticle.headline(), context)
        if match_value is None:
            raise Exception(("Could not access property `headline` of object with @id `" + s.Id) + "`")

        elif str(type(value(match_value))) == "<class \'str\'>":
            n: str = value(match_value)
            return n

        else: 
            raise Exception(("Value of property `headline` of object with @id `" + s.Id) + "` should have been a string")


    @staticmethod
    def set_headline_as_string(s: LDNode, n: str) -> None:
        s.SetProperty(LDScholarlyArticle.headline(), n)

    @staticmethod
    def get_identifiers(s: LDNode, context: LDContext | None=None) -> Array[Any]:
        return s.GetPropertyValues(LDScholarlyArticle.identifier(), None, context)

    @staticmethod
    def get_identifiers_as_property_value(s: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDPropertyValue.validate(ld_object, context_1)

        return s.GetPropertyNodes(LDScholarlyArticle.identifier(), filter, graph, context)

    @staticmethod
    def set_identifiers(s: LDNode, identifiers: Array[Any]) -> None:
        s.SetProperty(LDScholarlyArticle.identifier(), identifiers)

    @staticmethod
    def get_authors(s: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDPerson.validate(ld_object, context_1)

        return s.GetPropertyNodes(LDScholarlyArticle.author(), filter, graph, context)

    @staticmethod
    def set_authors(s: LDNode, authors: Array[LDNode], context: LDContext | None=None) -> Any:
        return s.SetProperty(LDScholarlyArticle.author(), authors, context)

    @staticmethod
    def try_get_url_as_string(s: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDScholarlyArticle.url(), context)
        (pattern_matching_result, u) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                u = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return u

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_same_as_as_string(s: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = s.TryGetPropertyAsSingleton(LDScholarlyArticle.same_as(), context)
        (pattern_matching_result, sa) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                sa = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return sa

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def try_get_creative_work_status(s: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = s.TryGetPropertyAsSingleNode(LDScholarlyArticle.creative_work_status(), graph, context)
        (pattern_matching_result, cws_1) = (None, None)
        if match_value is not None:
            def _arrow1506(__unit: None=None) -> bool:
                cws: LDNode = match_value
                return LDDefinedTerm.validate(cws, context)

            if _arrow1506():
                pattern_matching_result = 0
                cws_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return cws_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_creative_work_status(s: LDNode, cws: LDNode, context: LDContext | None=None) -> Any:
        return s.SetProperty(LDScholarlyArticle.creative_work_status(), cws, context)

    @staticmethod
    def get_comments(s: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDComment.validate(ld_object, context_1)

        return s.GetPropertyNodes(LDScholarlyArticle.comment(), filter, graph, context)

    @staticmethod
    def setcomments(s: LDNode, comments: Array[LDNode], context: LDContext | None=None) -> Any:
        return s.SetProperty(LDScholarlyArticle.comment(), comments, context)

    @staticmethod
    def gen_id(headline: str, url: str | None=None) -> str:
        return clean((("#" + headline) + "") if (url is None) else url)

    @staticmethod
    def validate(s: LDNode, context: LDContext | None=None) -> bool:
        return s.HasProperty(LDScholarlyArticle.headline(), context) if s.HasType(LDScholarlyArticle.schema_type(), context) else False

    @staticmethod
    def create(headline: str, identifiers: Array[Any], id: str | None=None, authors: Array[LDNode] | None=None, url: str | None=None, creative_work_status: LDNode | None=None, comments: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        s: LDNode = LDNode(LDScholarlyArticle.gen_id(headline, url) if (id is None) else id, [LDScholarlyArticle.schema_type()], None, context)
        s.SetProperty(LDScholarlyArticle.headline(), headline, context)
        s.SetProperty(LDScholarlyArticle.identifier(), identifiers, context)
        s.SetOptionalProperty(LDScholarlyArticle.author(), authors, context)
        s.SetOptionalProperty(LDScholarlyArticle.url(), url, context)
        s.SetOptionalProperty(LDScholarlyArticle.creative_work_status(), creative_work_status, context)
        s.SetOptionalProperty(LDScholarlyArticle.comment(), comments, context)
        return s


LDScholarlyArticle_reflection = _expr1507

__all__ = ["LDScholarlyArticle_reflection"]




from __future__ import annotations
from abc import abstractmethod
from openpyxl.worksheet.table import (TableStyleInfo, Table)
from typing import Protocol
from ..fable_openpyxl.openpyxl import Table as Table_1
from ..fs_spreadsheet.Ranges.fs_range_address import FsRangeAddress
from ..fs_spreadsheet.Tables.fs_table import FsTable

class tablestyle(Protocol):
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def show_column_stripes(self) -> bool:
        ...

    @abstractmethod
    def show_first_column(self) -> bool:
        ...

    @abstractmethod
    def show_last_column(self) -> bool:
        ...

    @abstractmethod
    def show_row_stripes(self) -> bool:
        ...


class TableStyleStatic(Protocol):
    @abstractmethod
    def create(self, name: str, showFirstColumn: bool, showLastColumn: bool, showRowStripes: bool, showColumnStripes: bool) -> tablestyle:
        ...


def default_table_style(__unit: None=None) -> tablestyle:
    return TableStyleInfo(name="TableStyleMedium9", showFirstColumn=False, showLastColumn=False, showRowStripes=True, showColumnStripes=True)


def from_fs_table(fs_table: FsTable) -> Table:
    table: Table = Table(displayName=fs_table.Name, ref=fs_table.RangeAddress.Range)
    table.tableStyleInfo = default_table_style()
    return table


def to_fs_table(table: Table) -> FsTable:
    name: str = table.name if ((table.displayName) is None) else (table.displayName)
    ref: str = table.ref
    return FsTable(name, FsRangeAddress.from_string(ref))


__all__ = ["default_table_style", "from_fs_table", "to_fs_table"]




from __future__ import annotations
from typing import Any
from ..arctrl_file_system.path import combine_many
from ..arctrl_validation_packages.validation_packages_config import ValidationPackagesConfig
from ..arctrl_yaml.validation_packages_config import (ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toYamlString_Static_71136F3F, ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_fromYamlString_Static_Z721C83C5)
from ..fable_library.array_ import equals_with
from ..fable_library.types import Array
from .contract import (DTOType as DTOType_1, Contract, DTO)

ValidationPackagesConfigHelper_ConfigFilePath: str = combine_many([".arc", "validation_packages.yml"])

ValidationPackagesConfigHelper_ReadContract: Contract = Contract("READ", ValidationPackagesConfigHelper_ConfigFilePath, DTOType_1(9), None)

def ValidationPackagesConfigExtensions__007CValidationPackagesYamlPath_007C__007C(input: Array[str]) -> str | None:
    (pattern_matching_result,) = (None,)
    def _arrow3705(x: str, y: str, input: Any=input) -> bool:
        return x == y

    if (len(input) == 2) if (not equals_with(_arrow3705, input, None)) else False:
        if input[0] == ".arc":
            if input[1] == "validation_packages.yml":
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return combine_many(input)

    elif pattern_matching_result == 1:
        return None



def ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_ToCreateContract(this: ValidationPackagesConfig) -> Contract:
    return Contract.create_create(ValidationPackagesConfigHelper_ConfigFilePath, DTOType_1(9), DTO(1, ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toYamlString_Static_71136F3F()(this)))


def ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_ToDeleteContract(this: ValidationPackagesConfig) -> Contract:
    return Contract.create_delete(ValidationPackagesConfigHelper_ConfigFilePath)


def ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toDeleteContract_Static_724DAE55(config: ValidationPackagesConfig) -> Contract:
    return ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_ToDeleteContract(config)


def ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toCreateContract_Static_724DAE55(config: ValidationPackagesConfig) -> Contract:
    return ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_ToCreateContract(config)


def ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_tryFromReadContract_Static_7570923F(c: Contract) -> ValidationPackagesConfig | None:
    (pattern_matching_result, p_1, yaml_1) = (None, None, None)
    if c.Operation == "READ":
        if c.DTOType is not None:
            if c.DTOType.tag == 9:
                if c.DTO is not None:
                    if c.DTO.tag == 1:
                        def _arrow3706(__unit: None=None, c: Any=c) -> bool:
                            yaml: str = c.DTO.fields[0]
                            return c.Path == ValidationPackagesConfigHelper_ConfigFilePath

                        if _arrow3706():
                            pattern_matching_result = 0
                            p_1 = c.Path
                            yaml_1 = c.DTO.fields[0]

                        else: 
                            pattern_matching_result = 1


                    else: 
                        pattern_matching_result = 1


                else: 
                    pattern_matching_result = 1


            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_fromYamlString_Static_Z721C83C5(yaml_1)

    elif pattern_matching_result == 1:
        return None



__all__ = ["ValidationPackagesConfigHelper_ConfigFilePath", "ValidationPackagesConfigHelper_ReadContract", "ValidationPackagesConfigExtensions__007CValidationPackagesYamlPath_007C__007C", "ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_ToCreateContract", "ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_ToDeleteContract", "ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toDeleteContract_Static_724DAE55", "ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_toCreateContract_Static_724DAE55", "ARCtrl_ValidationPackages_ValidationPackagesConfig__ValidationPackagesConfig_tryFromReadContract_Static_7570923F"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.Helper.identifier import create_missing_identifier
from ...fable_library.list import FSharpList
from ...fable_library.option import value
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import Array
from ..helper import clean
from ..ldcontext import LDContext
from ..ldobject import (LDNode, LDGraph)
from .comment import LDComment
from .organization import LDOrganization
from .person import LDPerson

def _expr1538() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.LDSoftwareSourceCode", None, LDSoftwareSourceCode)


class LDSoftwareSourceCode:
    @staticmethod
    def schema_type() -> str:
        return "http://schema.org/SoftwareSourceCode"

    @staticmethod
    def creator() -> str:
        return "http://schema.org/creator"

    @staticmethod
    def date_created() -> str:
        return "http://schema.org/dateCreated"

    @staticmethod
    def license() -> str:
        return "http://schema.org/license"

    @staticmethod
    def name() -> str:
        return "http://schema.org/name"

    @staticmethod
    def programming_language() -> str:
        return "http://schema.org/programmingLanguage"

    @staticmethod
    def sd_publisher() -> str:
        return "http://schema.org/sdPublisher"

    @staticmethod
    def url() -> str:
        return "http://schema.org/url"

    @staticmethod
    def version() -> str:
        return "http://schema.org/version"

    @staticmethod
    def description() -> str:
        return "http://schema.org/description"

    @staticmethod
    def has_part() -> str:
        return "http://schema.org/hasPart"

    @staticmethod
    def comment() -> str:
        return "http://schema.org/comment"

    @staticmethod
    def get_creator(ssc: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            if LDPerson.validate(ld_object, context_1):
                return True

            else: 
                return LDOrganization.validate(ld_object, context_1)


        return ssc.GetPropertyNodes(LDSoftwareSourceCode.creator(), filter, graph, context)

    @staticmethod
    def set_creator(ssc: LDNode, creator: LDNode, context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.creator(), creator, context)

    @staticmethod
    def try_get_date_created(ssc: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = ssc.TryGetPropertyAsSingleton(LDSoftwareSourceCode.date_created(), context)
        (pattern_matching_result, d) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                d = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return d

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_date_created(ssc: LDNode, date: str, context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.date_created(), date, context)

    @staticmethod
    def get_licenses(ssc: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return ssc.GetPropertyNodes(LDSoftwareSourceCode.license(), None, graph, context)

    @staticmethod
    def set_licenses(ssc: LDNode, licenses: Array[LDNode], context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.license(), licenses, context)

    @staticmethod
    def try_get_name_as_string(ssc: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = ssc.TryGetPropertyAsSingleton(LDSoftwareSourceCode.name(), context)
        (pattern_matching_result, n) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                n = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return n

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_name_as_string(ssc: LDNode, name: str, context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.name(), name, context)

    @staticmethod
    def get_programming_languages(ssc: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return ssc.GetPropertyNodes(LDSoftwareSourceCode.programming_language(), None, graph, context)

    @staticmethod
    def set_programming_languages(ssc: LDNode, programming_languages: Array[LDNode], context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.programming_language(), programming_languages, context)

    @staticmethod
    def try_get_sd_publisher(ssc: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> LDNode | None:
        match_value: LDNode | None = ssc.TryGetPropertyAsSingleNode(LDSoftwareSourceCode.sd_publisher(), graph, context)
        (pattern_matching_result, a_1) = (None, None)
        if match_value is not None:
            def _arrow1537(__unit: None=None) -> bool:
                ld_object: LDNode = match_value
                context_1: LDContext | None = context
                return True if LDPerson.validate(ld_object, context_1) else LDOrganization.validate(ld_object, context_1)

            if _arrow1537():
                pattern_matching_result = 0
                a_1 = match_value

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return a_1

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_sd_publisher(ssc: LDNode, sd_publisher: LDNode, context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.sd_publisher(), sd_publisher, context)

    @staticmethod
    def try_get_url(ssc: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = ssc.TryGetPropertyAsSingleton(LDSoftwareSourceCode.url(), context)
        (pattern_matching_result, u) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                u = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return u

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_url(ssc: LDNode, url: str, context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.url(), url, context)

    @staticmethod
    def try_get_version(ssc: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = ssc.TryGetPropertyAsSingleton(LDSoftwareSourceCode.version(), context)
        (pattern_matching_result, v) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                v = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return v

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_version(ssc: LDNode, version: str, context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.version(), version, context)

    @staticmethod
    def try_get_description(ssc: LDNode, context: LDContext | None=None) -> str | None:
        match_value: Any | None = ssc.TryGetPropertyAsSingleton(LDSoftwareSourceCode.description(), context)
        (pattern_matching_result, d) = (None, None)
        if match_value is not None:
            if str(type(value(match_value))) == "<class \'str\'>":
                pattern_matching_result = 0
                d = value(match_value)

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return d

        elif pattern_matching_result == 1:
            return None


    @staticmethod
    def set_description(ssc: LDNode, description: str, context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.description(), description, context)

    @staticmethod
    def get_has_part(ssc: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        return ssc.GetPropertyNodes(LDSoftwareSourceCode.has_part(), None, graph, context)

    @staticmethod
    def set_has_part(ssc: LDNode, has_parts: FSharpList[str], context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.has_part(), list(has_parts), context)

    @staticmethod
    def get_comments(ssc: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Array[LDNode]:
        def filter(ld_object: LDNode, context_1: LDContext | None=None) -> bool:
            return LDComment.validate(ld_object, context_1)

        return ssc.GetPropertyNodes(LDSoftwareSourceCode.comment(), filter, graph, context)

    @staticmethod
    def set_comments(ssc: LDNode, comments: Array[LDNode], context: LDContext | None=None) -> Any:
        return ssc.SetProperty(LDSoftwareSourceCode.comment(), comments, context)

    @staticmethod
    def validate(lp: LDNode, context: LDContext | None=None) -> bool:
        return lp.HasType(LDSoftwareSourceCode.schema_type(), context)

    @staticmethod
    def create(id: str | None=None, creator: LDNode | None=None, date_created: str | None=None, licenses: Array[LDNode] | None=None, name: str | None=None, programming_languages: Array[LDNode] | None=None, sd_publisher: LDNode | None=None, url: str | None=None, version: str | None=None, description: str | None=None, has_parts: Array[LDNode] | None=None, comments: Array[LDNode] | None=None, context: LDContext | None=None) -> LDNode:
        ssc: LDNode = LDNode(clean(("#ComputationalWorkflow_" + create_missing_identifier()) + "") if (id is None) else id, [LDSoftwareSourceCode.schema_type()], None, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.creator(), creator, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.date_created(), date_created, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.license(), licenses, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.name(), name, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.programming_language(), programming_languages, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.sd_publisher(), sd_publisher, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.url(), url, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.version(), version, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.description(), description, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.has_part(), has_parts, context)
        ssc.SetOptionalProperty(LDSoftwareSourceCode.comment(), comments, context)
        return ssc


LDSoftwareSourceCode_reflection = _expr1538

__all__ = ["LDSoftwareSourceCode_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...fable_library.array_ import map
from ...fable_library.reflection import (name, get_union_cases, TypeInfo, string_type, union_type)
from ...fable_library.reg_exp import (get_item, groups)
from ...fable_library.string_ import (to_fail, printf)
from ...fable_library.types import (Array, to_string, Union)
from ..Helper.regex import (try_parse_iotype_header, ActivePatterns__007CRegex_007C__007C, Pattern_InputPattern, Pattern_OutputPattern, Pattern_CommentPattern, ActivePatterns__007CTermColumn_007C__007C)
from ..ontology_annotation import (OntologyAnnotation, OntologyAnnotation_reflection)

__A = TypeVar("__A")

def _expr818() -> TypeInfo:
    return union_type("ARCtrl.IOType", [], IOType, lambda: [[], [], [], [], [("Item", string_type)]])


class IOType(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Source", "Sample", "Data", "Material", "FreeText"]

    @staticmethod
    def All() -> Array[IOType]:
        return [IOType(0), IOType(1), IOType(2), IOType(3)]

    @staticmethod
    def Cases() -> Array[tuple[int, str]]:
        def mapping(x: Any) -> tuple[int, str]:
            return (x.tag, name(x))

        return map(mapping, get_union_cases(IOType_reflection()), None)

    @property
    def as_input(self, __unit: None=None) -> str:
        this: IOType = self
        def string_create(x: Any | None=None) -> str:
            def _arrow813(__unit: None=None, x: Any=x) -> str:
                copy_of_struct: __A = x
                return to_string(copy_of_struct)

            return ("Input [" + _arrow813()) + "]"

        return string_create(this.fields[0]) if (this.tag == 4) else string_create(this)

    @property
    def as_output(self, __unit: None=None) -> str:
        this: IOType = self
        def string_create(x: Any | None=None) -> str:
            def _arrow814(__unit: None=None, x: Any=x) -> str:
                copy_of_struct: __A = x
                return to_string(copy_of_struct)

            return ("Output [" + _arrow814()) + "]"

        return string_create(this.fields[0]) if (this.tag == 4) else string_create(this)

    def Merge(self, other: IOType) -> IOType:
        this: IOType = self
        if this.tag == 2:
            if other.tag == 0:
                return IOType(2)

            else: 
                raise Exception(("Data IO column and " + str(other)) + " can not be merged")


        elif this.tag == 1:
            if other.tag == 0:
                return IOType(1)

            elif other.tag == 1:
                return IOType(1)

            else: 
                raise Exception(("Sample IO column and " + str(other)) + " can not be merged")


        elif this.tag == 0:
            return IOType(0) if (other.tag == 0) else other

        elif this.tag == 3:
            if other.tag == 0:
                return IOType(3)

            elif other.tag == 3:
                return IOType(3)

            else: 
                raise Exception(("Material IO column and " + str(other)) + " can not be merged")


        elif other.tag == 4:
            if this.fields[0] == other.fields[0]:
                return IOType(4, this.fields[0])

            else: 
                raise Exception(((("FreeText IO column names " + this.fields[0]) + " and ") + other.fields[0]) + " do differ")


        else: 
            raise Exception(("FreeText IO column and " + str(other)) + " can not be merged")


    def __str__(self, __unit: None=None) -> str:
        this: IOType = self
        if this.tag == 1:
            return "Sample Name"

        elif this.tag == 2:
            return "Data"

        elif this.tag == 3:
            return "Material"

        elif this.tag == 4:
            return this.fields[0]

        else: 
            return "Source Name"


    @staticmethod
    def of_string(str_1: str) -> IOType:
        return IOType(0) if (str_1 == "Source") else (IOType(0) if (str_1 == "Source Name") else (IOType(1) if (str_1 == "Sample") else (IOType(1) if (str_1 == "Sample Name") else (IOType(2) if (str_1 == "RawDataFile") else (IOType(2) if (str_1 == "Raw Data File") else (IOType(2) if (str_1 == "DerivedDataFile") else (IOType(2) if (str_1 == "Derived Data File") else (IOType(2) if (str_1 == "ImageFile") else (IOType(2) if (str_1 == "Image File") else (IOType(2) if (str_1 == "Data") else (IOType(3) if (str_1 == "Material") else IOType(4, str_1))))))))))))

    @staticmethod
    def try_of_header_string(str_1: str) -> IOType | None:
        match_value: str | None = try_parse_iotype_header(str_1)
        if match_value is None:
            return None

        else: 
            s: str = match_value
            return IOType.of_string(s)


    @staticmethod
    def source(__unit: None=None) -> IOType:
        return IOType(0)

    @staticmethod
    def sample(__unit: None=None) -> IOType:
        return IOType(1)

    @staticmethod
    def data(__unit: None=None) -> IOType:
        return IOType(2)

    @staticmethod
    def material(__unit: None=None) -> IOType:
        return IOType(3)

    @staticmethod
    def free_text(s: str) -> IOType:
        return IOType(4, s)


IOType_reflection = _expr818

def _expr820() -> TypeInfo:
    return union_type("ARCtrl.CompositeHeader", [], CompositeHeader, lambda: [[("Item", OntologyAnnotation_reflection())], [("Item", OntologyAnnotation_reflection())], [("Item", OntologyAnnotation_reflection())], [("Item", OntologyAnnotation_reflection())], [], [], [], [], [], [], [], [("Item", IOType_reflection())], [("Item", IOType_reflection())], [("Item", string_type)], [("Item", string_type)]])


class CompositeHeader(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Component", "Characteristic", "Factor", "Parameter", "ProtocolType", "ProtocolDescription", "ProtocolUri", "ProtocolVersion", "ProtocolREF", "Performer", "Date", "Input", "Output", "FreeText", "Comment"]

    @staticmethod
    def Cases() -> Array[tuple[int, str]]:
        def mapping(x: Any) -> tuple[int, str]:
            return (x.tag, name(x))

        return map(mapping, get_union_cases(CompositeHeader_reflection()), None)

    @staticmethod
    def js_get_column_meta_type(inp: int) -> int:
        return 1 if (inp == 0) else (1 if (inp == 1) else (1 if (inp == 2) else (1 if (inp == 3) else (0 if (inp == 4) else (0 if (inp == 5) else (0 if (inp == 6) else (0 if (inp == 7) else (0 if (inp == 8) else (0 if (inp == 9) else (0 if (inp == 10) else (2 if (inp == 11) else (2 if (inp == 12) else (3 if (inp == 13) else (3 if (inp == 14) else to_fail(printf("Cannot assign input `Tag` (%i) to `CompositeHeader`"))(inp)))))))))))))))

    def __str__(self, __unit: None=None) -> str:
        this: CompositeHeader = self
        if this.tag == 2:
            return ("Factor [" + this.fields[0].NameText) + "]"

        elif this.tag == 1:
            return ("Characteristic [" + this.fields[0].NameText) + "]"

        elif this.tag == 0:
            return ("Component [" + this.fields[0].NameText) + "]"

        elif this.tag == 4:
            return "Protocol Type"

        elif this.tag == 8:
            return "Protocol REF"

        elif this.tag == 5:
            return "Protocol Description"

        elif this.tag == 6:
            return "Protocol Uri"

        elif this.tag == 7:
            return "Protocol Version"

        elif this.tag == 9:
            return "Performer"

        elif this.tag == 10:
            return "Date"

        elif this.tag == 11:
            return this.fields[0].as_input

        elif this.tag == 12:
            return this.fields[0].as_output

        elif this.tag == 14:
            return ("Comment [" + this.fields[0]) + "]"

        elif this.tag == 13:
            return this.fields[0]

        else: 
            return ("Parameter [" + this.fields[0].NameText) + "]"


    def ToTerm(self, __unit: None=None) -> OntologyAnnotation:
        this: CompositeHeader = self
        if this.tag == 2:
            return this.fields[0]

        elif this.tag == 1:
            return this.fields[0]

        elif this.tag == 0:
            return this.fields[0]

        elif this.tag == 4:
            return OntologyAnnotation.create(to_string(this), None, this.GetFeaturedColumnAccession)

        elif this.tag == 8:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 5:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 6:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 7:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 9:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 10:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 11:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 12:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 14:
            return OntologyAnnotation.create(to_string(this))

        elif this.tag == 13:
            return OntologyAnnotation.create(to_string(this))

        else: 
            return this.fields[0]


    @staticmethod
    def OfHeaderString(str_1: str) -> CompositeHeader:
        match_value: str = str_1.strip()
        active_pattern_result: Any | None = ActivePatterns__007CRegex_007C__007C(Pattern_InputPattern, match_value)
        if active_pattern_result is not None:
            r: Any = active_pattern_result
            iotype: str = get_item(groups(r), "iotype") or ""
            return CompositeHeader(11, IOType.of_string(iotype))

        else: 
            active_pattern_result_1: Any | None = ActivePatterns__007CRegex_007C__007C(Pattern_OutputPattern, match_value)
            if active_pattern_result_1 is not None:
                r_1: Any = active_pattern_result_1
                iotype_1: str = get_item(groups(r_1), "iotype") or ""
                return CompositeHeader(12, IOType.of_string(iotype_1))

            else: 
                active_pattern_result_2: Any | None = ActivePatterns__007CRegex_007C__007C(Pattern_CommentPattern, match_value)
                if active_pattern_result_2 is not None:
                    r_2: Any = active_pattern_result_2
                    return CompositeHeader(14, get_item(groups(r_2), "commentKey") or "")

                else: 
                    active_pattern_result_3: dict[str, Any] | None = ActivePatterns__007CTermColumn_007C__007C(match_value)
                    if active_pattern_result_3 is not None:
                        r_3: dict[str, Any] = active_pattern_result_3
                        match_value_1: str = r_3["TermColumnType"]
                        (pattern_matching_result,) = (None,)
                        if match_value_1 == "Parameter":
                            pattern_matching_result = 0

                        elif match_value_1 == "Parameter Value":
                            pattern_matching_result = 0

                        elif match_value_1 == "Factor":
                            pattern_matching_result = 1

                        elif match_value_1 == "Factor Value":
                            pattern_matching_result = 1

                        elif match_value_1 == "Characteristic":
                            pattern_matching_result = 2

                        elif match_value_1 == "Characteristics":
                            pattern_matching_result = 2

                        elif match_value_1 == "Characteristics Value":
                            pattern_matching_result = 2

                        elif match_value_1 == "Component":
                            pattern_matching_result = 3

                        else: 
                            pattern_matching_result = 4

                        if pattern_matching_result == 0:
                            return CompositeHeader(3, OntologyAnnotation.create(r_3["TermName"]))

                        elif pattern_matching_result == 1:
                            return CompositeHeader(2, OntologyAnnotation.create(r_3["TermName"]))

                        elif pattern_matching_result == 2:
                            return CompositeHeader(1, OntologyAnnotation.create(r_3["TermName"]))

                        elif pattern_matching_result == 3:
                            return CompositeHeader(0, OntologyAnnotation.create(r_3["TermName"]))

                        elif pattern_matching_result == 4:
                            return CompositeHeader(13, str_1)


                    else: 
                        return CompositeHeader(10) if (match_value == "Date") else (CompositeHeader(9) if (match_value == "Performer") else (CompositeHeader(5) if (match_value == "Protocol Description") else (CompositeHeader(6) if (match_value == "Protocol Uri") else (CompositeHeader(7) if (match_value == "Protocol Version") else (CompositeHeader(4) if (match_value == "Protocol Type") else (CompositeHeader(8) if (match_value == "Protocol REF") else CompositeHeader(13, match_value)))))))





    @property
    def IsDeprecated(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        (pattern_matching_result,) = (None,)
        if this.tag == 13:
            if this.fields[0].lower() == "sample name":
                pattern_matching_result = 0

            elif this.fields[0].lower() == "source name":
                pattern_matching_result = 1

            elif this.fields[0].lower() == "data file name":
                pattern_matching_result = 2

            elif this.fields[0].lower() == "derived data file":
                pattern_matching_result = 3

            else: 
                pattern_matching_result = 4


        else: 
            pattern_matching_result = 4

        if pattern_matching_result == 0:
            return True

        elif pattern_matching_result == 1:
            return True

        elif pattern_matching_result == 2:
            return True

        elif pattern_matching_result == 3:
            return True

        elif pattern_matching_result == 4:
            return False


    @property
    def IsCvParamColumn(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        if (((this.tag == 3) or (this.tag == 2)) or (this.tag == 1)) or (this.tag == 0):
            return True

        else: 
            return False


    @property
    def IsTermColumn(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        if ((((this.tag == 3) or (this.tag == 2)) or (this.tag == 1)) or (this.tag == 0)) or (this.tag == 4):
            return True

        else: 
            return False


    @property
    def IsDataColumn(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        (pattern_matching_result,) = (None,)
        if this.tag == 11:
            if this.fields[0].tag == 2:
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        elif this.tag == 12:
            if this.fields[0].tag == 2:
                pattern_matching_result = 0

            else: 
                pattern_matching_result = 1


        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return True

        elif pattern_matching_result == 1:
            return False


    @property
    def IsFeaturedColumn(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 4) else False

    @property
    def GetFeaturedColumnAccession(self, __unit: None=None) -> str:
        this: CompositeHeader = self
        if this.tag == 4:
            return "DPBO:1000161"

        else: 
            raise Exception(("Tried matching " + str(this)) + " in getFeaturedColumnAccession, but is not a featured column.")


    @property
    def GetColumnAccessionShort(self, __unit: None=None) -> str:
        this: CompositeHeader = self
        if this.tag == 4:
            return this.GetFeaturedColumnAccession

        elif this.tag == 3:
            return this.fields[0].TermAccessionShort

        elif this.tag == 2:
            return this.fields[0].TermAccessionShort

        elif this.tag == 1:
            return this.fields[0].TermAccessionShort

        elif this.tag == 0:
            return this.fields[0].TermAccessionShort

        else: 
            raise Exception(("Tried matching " + str(this)) + ", but is not a column with an accession.")


    @property
    def IsSingleColumn(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        if (((((((((this.tag == 13) or (this.tag == 11)) or (this.tag == 12)) or (this.tag == 14)) or (this.tag == 8)) or (this.tag == 5)) or (this.tag == 6)) or (this.tag == 7)) or (this.tag == 9)) or (this.tag == 10):
            return True

        else: 
            return False


    @property
    def IsIOType(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        if this.tag == 11:
            return True

        elif this.tag == 12:
            return True

        else: 
            return False


    @property
    def is_input(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 11) else False

    @property
    def is_output(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 12) else False

    @property
    def is_parameter(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 3) else False

    @property
    def is_factor(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 2) else False

    @property
    def is_characteristic(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 1) else False

    @property
    def is_component(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 0) else False

    @property
    def is_protocol_type(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 4) else False

    @property
    def is_protocol_ref(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 8) else False

    @property
    def is_protocol_description(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 5) else False

    @property
    def is_protocol_uri(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 6) else False

    @property
    def is_protocol_version(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 7) else False

    @property
    def is_protocol_column(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        if ((((this.tag == 8) or (this.tag == 5)) or (this.tag == 6)) or (this.tag == 7)) or (this.tag == 4):
            return True

        else: 
            return False


    @property
    def is_performer(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 9) else False

    @property
    def is_date(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 10) else False

    @property
    def is_comment(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 14) else False

    @property
    def is_free_text(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        return True if (this.tag == 13) else False

    def TryInput(self, __unit: None=None) -> IOType | None:
        this: CompositeHeader = self
        return this.fields[0] if (this.tag == 11) else None

    def TryOutput(self, __unit: None=None) -> IOType | None:
        this: CompositeHeader = self
        return this.fields[0] if (this.tag == 12) else None

    def TryIOType(self, __unit: None=None) -> IOType | None:
        this: CompositeHeader = self
        (pattern_matching_result, io) = (None, None)
        if this.tag == 12:
            pattern_matching_result = 0
            io = this.fields[0]

        elif this.tag == 11:
            pattern_matching_result = 0
            io = this.fields[0]

        else: 
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return io

        elif pattern_matching_result == 1:
            return None


    @property
    def IsUnique(self, __unit: None=None) -> bool:
        this: CompositeHeader = self
        if ((((((((this.tag == 4) or (this.tag == 8)) or (this.tag == 5)) or (this.tag == 6)) or (this.tag == 7)) or (this.tag == 9)) or (this.tag == 10)) or (this.tag == 11)) or (this.tag == 12):
            return True

        else: 
            return False


    def Copy(self, __unit: None=None) -> CompositeHeader:
        this: CompositeHeader = self
        if this.tag == 3:
            return CompositeHeader(3, this.fields[0].Copy())

        elif this.tag == 2:
            return CompositeHeader(2, this.fields[0].Copy())

        elif this.tag == 1:
            return CompositeHeader(1, this.fields[0].Copy())

        elif this.tag == 0:
            return CompositeHeader(0, this.fields[0].Copy())

        else: 
            return this


    def TryGetTerm(self, __unit: None=None) -> OntologyAnnotation | None:
        this: CompositeHeader = self
        if this.tag == 3:
            return this.fields[0]

        elif this.tag == 2:
            return this.fields[0]

        elif this.tag == 1:
            return this.fields[0]

        elif this.tag == 0:
            return this.fields[0]

        else: 
            return None


    @staticmethod
    def component(oa: OntologyAnnotation) -> CompositeHeader:
        return CompositeHeader(0, oa)

    @staticmethod
    def characteristic(oa: OntologyAnnotation) -> CompositeHeader:
        return CompositeHeader(1, oa)

    @staticmethod
    def factor(oa: OntologyAnnotation) -> CompositeHeader:
        return CompositeHeader(2, oa)

    @staticmethod
    def parameter(oa: OntologyAnnotation) -> CompositeHeader:
        return CompositeHeader(3, oa)

    @staticmethod
    def protocol_type(__unit: None=None) -> CompositeHeader:
        return CompositeHeader(4)

    @staticmethod
    def protocol_description(__unit: None=None) -> CompositeHeader:
        return CompositeHeader(5)

    @staticmethod
    def protocol_uri(__unit: None=None) -> CompositeHeader:
        return CompositeHeader(6)

    @staticmethod
    def protocol_version(__unit: None=None) -> CompositeHeader:
        return CompositeHeader(7)

    @staticmethod
    def protocol_ref(__unit: None=None) -> CompositeHeader:
        return CompositeHeader(8)

    @staticmethod
    def performer(__unit: None=None) -> CompositeHeader:
        return CompositeHeader(9)

    @staticmethod
    def date(__unit: None=None) -> CompositeHeader:
        return CompositeHeader(10)

    @staticmethod
    def input(io: IOType) -> CompositeHeader:
        return CompositeHeader(11, io)

    @staticmethod
    def output(io: IOType) -> CompositeHeader:
        return CompositeHeader(12, io)

    @staticmethod
    def free_text(s: str) -> CompositeHeader:
        return CompositeHeader(13, s)

    @staticmethod
    def comment(s: str) -> CompositeHeader:
        return CompositeHeader(14, s)


CompositeHeader_reflection = _expr820

__all__ = ["IOType_reflection", "CompositeHeader_reflection"]




from __future__ import annotations

from typing import Any, Generic, TypeVar

from .option import some
from .reflection import TypeInfo, union_type
from .types import Array, Union


_T1 = TypeVar("_T1")

_T2 = TypeVar("_T2")

_T3 = TypeVar("_T3")

_T4 = TypeVar("_T4")

_T5 = TypeVar("_T5")

_T6 = TypeVar("_T6")

_T7 = TypeVar("_T7")


def _expr35(gen0: TypeInfo, gen1: TypeInfo) -> TypeInfo:
    return union_type(
        "FSharp.Core.FSharpChoice`2", [gen0, gen1], FSharpChoice_2, lambda: [[("Item", gen0)], [("Item", gen1)]]
    )


class FSharpChoice_2(Union, Generic[_T1, _T2]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Choice1Of2", "Choice2Of2"]


FSharpChoice_2_reflection = _expr35


def _expr36(gen0: TypeInfo, gen1: TypeInfo, gen2: TypeInfo) -> TypeInfo:
    return union_type(
        "FSharp.Core.FSharpChoice`3",
        [gen0, gen1, gen2],
        FSharpChoice_3,
        lambda: [[("Item", gen0)], [("Item", gen1)], [("Item", gen2)]],
    )


class FSharpChoice_3(Union, Generic[_T1, _T2, _T3]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Choice1Of3", "Choice2Of3", "Choice3Of3"]


FSharpChoice_3_reflection = _expr36


def _expr37(gen0: TypeInfo, gen1: TypeInfo, gen2: TypeInfo, gen3: TypeInfo) -> TypeInfo:
    return union_type(
        "FSharp.Core.FSharpChoice`4",
        [gen0, gen1, gen2, gen3],
        FSharpChoice_4,
        lambda: [[("Item", gen0)], [("Item", gen1)], [("Item", gen2)], [("Item", gen3)]],
    )


class FSharpChoice_4(Union, Generic[_T1, _T2, _T3, _T4]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Choice1Of4", "Choice2Of4", "Choice3Of4", "Choice4Of4"]


FSharpChoice_4_reflection = _expr37


def _expr38(gen0: TypeInfo, gen1: TypeInfo, gen2: TypeInfo, gen3: TypeInfo, gen4: TypeInfo) -> TypeInfo:
    return union_type(
        "FSharp.Core.FSharpChoice`5",
        [gen0, gen1, gen2, gen3, gen4],
        FSharpChoice_5,
        lambda: [[("Item", gen0)], [("Item", gen1)], [("Item", gen2)], [("Item", gen3)], [("Item", gen4)]],
    )


class FSharpChoice_5(Union, Generic[_T1, _T2, _T3, _T4, _T5]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Choice1Of5", "Choice2Of5", "Choice3Of5", "Choice4Of5", "Choice5Of5"]


FSharpChoice_5_reflection = _expr38


def _expr39(gen0: TypeInfo, gen1: TypeInfo, gen2: TypeInfo, gen3: TypeInfo, gen4: TypeInfo, gen5: TypeInfo) -> TypeInfo:
    return union_type(
        "FSharp.Core.FSharpChoice`6",
        [gen0, gen1, gen2, gen3, gen4, gen5],
        FSharpChoice_6,
        lambda: [
            [("Item", gen0)],
            [("Item", gen1)],
            [("Item", gen2)],
            [("Item", gen3)],
            [("Item", gen4)],
            [("Item", gen5)],
        ],
    )


class FSharpChoice_6(Union, Generic[_T1, _T2, _T3, _T4, _T5, _T6]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Choice1Of6", "Choice2Of6", "Choice3Of6", "Choice4Of6", "Choice5Of6", "Choice6Of6"]


FSharpChoice_6_reflection = _expr39


def _expr40(
    gen0: TypeInfo, gen1: TypeInfo, gen2: TypeInfo, gen3: TypeInfo, gen4: TypeInfo, gen5: TypeInfo, gen6: TypeInfo
) -> TypeInfo:
    return union_type(
        "FSharp.Core.FSharpChoice`7",
        [gen0, gen1, gen2, gen3, gen4, gen5, gen6],
        FSharpChoice_7,
        lambda: [
            [("Item", gen0)],
            [("Item", gen1)],
            [("Item", gen2)],
            [("Item", gen3)],
            [("Item", gen4)],
            [("Item", gen5)],
            [("Item", gen6)],
        ],
    )


class FSharpChoice_7(Union, Generic[_T1, _T2, _T3, _T4, _T5, _T6, _T7]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Choice1Of7", "Choice2Of7", "Choice3Of7", "Choice4Of7", "Choice5Of7", "Choice6Of7", "Choice7Of7"]


FSharpChoice_7_reflection = _expr40


def Choice_makeChoice1Of2(x: Any | None = None) -> FSharpChoice_2[Any, Any]:
    return FSharpChoice_2(0, x)


def Choice_makeChoice2Of2(x: Any | None = None) -> FSharpChoice_2[Any, Any]:
    return FSharpChoice_2(1, x)


def Choice_tryValueIfChoice1Of2(x: FSharpChoice_2[Any, Any]) -> Any | None:
    if x.tag == 0:
        return some(x.fields[0])

    else:
        return None


def Choice_tryValueIfChoice2Of2(x: FSharpChoice_2[Any, Any]) -> Any | None:
    if x.tag == 1:
        return some(x.fields[0])

    else:
        return None


__all__ = [
    "FSharpChoice_2_reflection",
    "FSharpChoice_3_reflection",
    "FSharpChoice_4_reflection",
    "FSharpChoice_5_reflection",
    "FSharpChoice_6_reflection",
    "FSharpChoice_7_reflection",
    "Choice_makeChoice1Of2",
    "Choice_makeChoice2Of2",
    "Choice_tryValueIfChoice1Of2",
    "Choice_tryValueIfChoice2Of2",
]
